import asyncio
import collections
import logging
from typing import Dict, Any, List, Optional, TYPE_CHECKING, Tuple
from uuid import uuid4
import threading
from typing import Set

class RequiredFieldsTracker:
    def __init__(self):
        self._required_fields: Set[str] = set()
        self._lock = threading.Lock()
        self._event = threading.Event()

    def get_fields(self) -> Set[str]:
        with self._lock:
            return self._required_fields.copy()

    def update_fields(self, new_fields: Set[str]):
        with self._lock:
            if self._required_fields != new_fields:
                self._required_fields = new_fields
                self._event.set()

    def clear_event(self):
        self._event.clear()

    def wait_for_change(self, timeout: Optional[float] = None) -> bool:
        return self._event.wait(timeout)

from .base import BaseInstanceService
from fustor_core.models.states import EventBusInstance, EventBusState
from fustor_core.models.config import SourceConfig, FieldMapping
from fustor_agent.runtime.bus import MemoryEventBus
from fustor_core.exceptions import ConfigError, NotFoundError, DriverError, TransientSourceBufferFullError

if TYPE_CHECKING:
    from fustor_agent.services.instances.sync import SyncInstanceService
    from fustor_agent.services.drivers.source_driver import SourceDriverService

logger = logging.getLogger("fustor_agent")
from fustor_agent_sdk.interfaces import EventBusServiceInterface # Import the interface

class EventBusInstanceRuntime:
    def __init__(self, source_id, source_config: SourceConfig, source_signature: Any, source_driver_service: "SourceDriverService", initial_start_position: int = 0, loop: Optional[asyncio.AbstractEventLoop] = None):
        self.id: str = f"bus-{uuid4()}"
        self.source_id: str = source_id
        self.config: SourceConfig = source_config
        self.source_signature = source_signature
        
        source_driver_class = source_driver_service._get_driver_by_type(self.config.driver)
        self.source_driver_instance = source_driver_class(source_id, self.config)

        self.internal_bus: MemoryEventBus = MemoryEventBus(
            bus_id=self.id, 
            capacity=source_config.max_queue_size, 
            start_position=initial_start_position
        )
        
        self.state: EventBusState = EventBusState.IDLE
        self.info: str = "总线已创建，等待订阅者以启动数据生产。"
        self.statistics: Dict[str, Any] = {"events_produced": 0, "subscriber_count": 0}
        self.metrics: Dict[str, Any] = {"events_produced_total": 0, "bus_queue_size": 0}
        self.producer_task: Optional[asyncio.Task] = None
        self._producer_stop_event = threading.Event()
        self.required_fields_tracker = RequiredFieldsTracker()
        self.needed_position_lost: bool = False # Add new flag
        self._loop = loop or asyncio.get_running_loop()
        self.ready_event = asyncio.Event()

    def update_driver_required_fields(self, new_fields: Set[str]):
        self.required_fields_tracker.update_fields(new_fields)

    async def start_producer(self):
        if self.producer_task and not self.producer_task.done():
            logger.warning(f"总线 '{self.id}' 的生产者任务已在运行。")
            return

        self.state = EventBusState.PRODUCING
        self.info = "生产者任务已启动，正在监听数据源。"
        self._producer_stop_event.clear()
        self.producer_task = asyncio.create_task(self._produce_loop())
        logger.info(self.info)

    async def stop_producer(self):
        if self.producer_task and not self.producer_task.done():
            self._producer_stop_event.set()
            self.producer_task.cancel()
            try:
                await self.producer_task
            except asyncio.CancelledError:
                pass
        self.state = EventBusState.IDLE
        self.info = "因没有订阅者，生产者任务已暂停。"
        self.producer_task = None
        logger.info(self.info)

    async def _produce_loop(self):
        logger.info(f"总线 '{self.id}' 的生产循环已启动。")
        import queue
        event_queue = queue.Queue(maxsize=100)

        def _threaded_producer():
            logger.info(f"总线 '{self.id}' 的后台生产者线程已启动。")
            try:
                # Check if the requested position is available before getting the iterator
                start_position = self.internal_bus.buffer_start_position
                self.needed_position_lost = not self.source_driver_instance.is_position_available(start_position)
                
                # Get the iterator with the appropriate start position
                iterator = self.source_driver_instance.get_message_iterator(
                    start_position=start_position if not self.needed_position_lost else -1,
                    stop_event=self._producer_stop_event,
                    required_fields_tracker=self.required_fields_tracker
                )
                
                # Signal that the producer is ready (e.g., pre-scan is done and iterator is obtained)
                self._loop.call_soon_threadsafe(self.ready_event.set)
                logger.info(f"总线 '{self.id}' 的后台生产者线程已就绪，进入事件循环。")

                for event in iterator:
                    if self._producer_stop_event.is_set() or self.source_driver_instance._stop_driver_event.is_set(): # Check new stop event
                        break
                    event_queue.put(event)
            except DriverError as e: # Catch the specific DriverError
                logger.error(f"总线 '{self.id}' 的后台生产者因驱动错误而停止: {e}", exc_info=True)
                event_queue.put(e) # Propagate the error
            except Exception as e:
                logger.error(f"总线 '{self.id}' 的后台生产者出现异常: {e}", exc_info=True)
                event_queue.put(e)
            finally:
                # Always put None to unblock the consumer, regardless of how the loop exited.
                if not self.ready_event.is_set():
                    self._loop.call_soon_threadsafe(self.ready_event.set)
                event_queue.put(None)
                logger.info(f"总线 '{self.id}' 的后台生产者线程已结束。")

        producer_thread = threading.Thread(target=_threaded_producer, daemon=True)
        producer_thread.start()

        try:
            while True:
                event = await asyncio.to_thread(event_queue.get)

                if event is None:
                    logger.info(f"总线 '{self.id}' 的生产者已发出完成信号。")
                    break
                
                if isinstance(event, Exception):
                    if isinstance(event, DriverError):
                        error_msg = f"总线 '{self.id}' 的生产循环因驱动错误而停止: {event}"
                        self.state = EventBusState.ERROR
                        self.info = error_msg
                        logger.error(error_msg, exc_info=True)
                        self.internal_bus.mark_as_failed(str(event))
                        break # Stop the loop on DriverError
                    else:
                        raise event

                await self.internal_bus.put(event, is_transient=self.source_driver_instance.is_transient)
                if hasattr(event, 'rows'):
                    self.statistics["events_produced"] += len(event.rows)
                    self.metrics["events_produced_total"] += len(event.rows)
                self.metrics["bus_queue_size"] = len(self.internal_bus.buffer)

        except asyncio.CancelledError:
            logger.info(f"总线 '{self.id}' 的生产循环被取消。")
        except TransientSourceBufferFullError as e:
            error_msg = str(e)
            self.state = EventBusState.ERROR
            self.info = error_msg
            logger.error(f"总线 '{self.id}' 因缓冲区满而停止: {error_msg}")
            self.internal_bus.mark_as_failed(error_msg)
        except Exception as e:
            error_msg = f"总线 '{self.id}' 的生产循环崩溃: {e}"
            self.state = EventBusState.ERROR
            self.info = error_msg
            logger.error(error_msg, exc_info=True)
            self.internal_bus.mark_as_failed(str(e))
        finally:
            self._producer_stop_event.set()
            if producer_thread.is_alive():
                await asyncio.to_thread(producer_thread.join, timeout=5.0)
            logger.info(f"总线 '{self.id}' 的生产循环已结束。")

    def get_dto(self) -> EventBusInstance:
        self.statistics['buffer_size'] = len(self.internal_bus.buffer)
        self.statistics['subscriber_count'] = self.internal_bus.get_subscriber_count()
        self.statistics.update(self.metrics)
        return EventBusInstance(
            id=self.id,
            source_name=self.source_id,
            state=self.state,
            info=self.info,
            statistics=self.statistics
        )

class EventBusService(BaseInstanceService, EventBusServiceInterface): # Inherit from the interface
    def __init__(self, source_configs: Dict[str, SourceConfig], source_driver_service: "SourceDriverService"):
        super().__init__()
        self.source_configs = source_configs
        self.source_driver_service = source_driver_service
        self.bus_by_signature: Dict[Any, EventBusInstanceRuntime] = {}
        self._bus_creation_locks: Dict[str, asyncio.Lock] = collections.defaultdict(asyncio.Lock)

    def set_dependencies(self, sync_instance_service: "SyncInstanceService"):
        self.sync_instance_service = sync_instance_service

    def _generate_source_signature(self, source_config: SourceConfig) -> Any:
        return (source_config.driver, source_config.uri, source_config.credential)

    async def get_or_create_bus_for_subscriber(
        self, 
        source_id: str,
        source_config: SourceConfig, 
        sync_id: str,
        required_position: int,
        fields_mapping: List[FieldMapping]
    ) -> Tuple[EventBusInstanceRuntime, bool]: # Updated return type
        source_signature = self._generate_source_signature(source_config)
        
        async with self._bus_creation_locks[str(source_signature)]:
            bus_runtime = self.bus_by_signature.get(source_signature)

            if bus_runtime and bus_runtime.state == EventBusState.ERROR:
                logger.warning(f"Removed failed EventBus '{bus_runtime.id}' for source '{source_id}'. Creating a new one.")
                self.pool.pop(bus_runtime.id, None)
                bus_runtime = None

            if bus_runtime and not bus_runtime.internal_bus.can_subscribe(required_position):
                logger.info(f"Existing bus '{bus_runtime.id}' is unsuitable for sync '{sync_id}' (requires position {required_position}, bus starts at {bus_runtime.internal_bus.buffer_start_position}). Creating a new bus.")
                bus_runtime = None

            if not bus_runtime:
                bus_runtime = EventBusInstanceRuntime(
                    source_id=source_id,
                    source_config=source_config, 
                    source_signature=source_signature, 
                    source_driver_service=self.source_driver_service,
                    initial_start_position=required_position,
                    loop=asyncio.get_running_loop()
                )
                self.pool[bus_runtime.id] = bus_runtime
                self.bus_by_signature[source_signature] = bus_runtime

        # Ensure the producer is running for the bus we are about to use.
        # This is crucial for recovered buses that don't have an active producer task.
        if not bus_runtime.producer_task or bus_runtime.producer_task.done():
            await bus_runtime.start_producer()

        # All callers must wait for the producer to be ready before proceeding
        await bus_runtime.ready_event.wait()

        await bus_runtime.internal_bus.subscribe(sync_id, required_position, fields_mapping)
        bus_runtime.update_driver_required_fields(bus_runtime.internal_bus.required_fields)
        
        # Return the tuple with the signal
        return bus_runtime, bus_runtime.needed_position_lost

    async def release_subscriber(self, bus_id: str, sync_id: str):
        bus_runtime = self.get_instance(bus_id)
        if not bus_runtime:
            logger.warning(f"Attempted to release subscriber from non-existent bus '{bus_id}'.")
            return
        
        await bus_runtime.internal_bus.unsubscribe(sync_id)
        
        if bus_runtime.internal_bus.get_subscriber_count() == 0:
            logger.info(f"Bus '{bus_id}' has no more subscribers, stopping its producer.")
            await bus_runtime.stop_producer()
            self.pool.pop(bus_id, None)
            if bus_runtime.source_signature in self.bus_by_signature:
                del self.bus_by_signature[bus_runtime.source_signature]

    async def release_all_unused_buses(self):
        logger.info("Releasing all unused event buses...")
        for bus_id in list(self.pool.keys()):
            bus = self.pool.get(bus_id)
            if bus and bus.internal_bus.get_subscriber_count() == 0:
                await bus.stop_producer()
                self.pool.pop(bus_id, None)
                if bus.source_signature in self.bus_by_signature:
                    del self.bus_by_signature[bus.source_signature]

    async def commit_and_handle_split(
        self, 
        bus_id: str, 
        sync_id: str, 
        num_events: int, 
        last_consumed_position: int,
        fields_mapping: List[FieldMapping]
    ):
        bus_runtime = self.get_instance(bus_id)
        if not bus_runtime:
            return

        task_to_split_id = await bus_runtime.internal_bus.commit(sync_id, num_events, last_consumed_position)

        if task_to_split_id:
            logger.warning(f"Bus split condition met on '{bus_id}'. Task '{task_to_split_id}' will be migrated.")
            
            source_config = self.source_configs.get(bus_runtime.source_id)
            if not source_config:
                logger.error(f"Cannot split task: Source config '{bus_runtime.source_id}' not found.")
                return

            # When splitting, the new bus must also return the signal
            new_bus, needed_position_lost = await self.get_or_create_bus_for_subscriber(
                source_id=bus_runtime.source_id,
                source_config=source_config,
                sync_id=task_to_split_id,
                required_position=last_consumed_position,
                fields_mapping=fields_mapping
            )
            
            # The remapping sync instance needs to know about the signal too
            await self.sync_instance_service.remap_sync_to_new_bus(task_to_split_id, new_bus, needed_position_lost)


import pytest
from unittest.mock import MagicMock, AsyncMock
from fustor_agent.services.drivers.pusher_driver import PusherDriverService
from fustor_core.exceptions import DriverError, ConfigError

@pytest.fixture
def pusher_driver_service(mocker):
    """Fixture for a PusherDriverService with patched discovery."""
    mocker.patch.object(PusherDriverService, '_discover_installed_drivers', return_value={})
    yield PusherDriverService()

@pytest.fixture
def mock_pusher_class():
    """Mocks a driver class that conforms to the PusherDriver ABC."""
    mock = MagicMock()
    # Mock class methods
    mock.get_wizard_steps = AsyncMock(return_value={"steps": [{"id": "s1"}]})
    mock.get_needed_fields = AsyncMock(return_value={"properties": {"field1": {"type": "string"}}})
    mock.test_connection = AsyncMock(return_value=(True, "Connected"))
    mock.check_privileges = AsyncMock(return_value=(True, "Privileges OK"))
    return mock

class TestPusherDriverService:

    def test_get_driver_by_type_success(self, pusher_driver_service, mock_pusher_class, mocker):
        """Tests that a discovered driver class can be retrieved."""
        mocker.patch.dict(pusher_driver_service._discovered_drivers, {"mock_pusher": mock_pusher_class})
        driver_class = pusher_driver_service._get_driver_by_type("mock_pusher")
        assert driver_class == mock_pusher_class

    def test_get_driver_by_type_not_found(self, pusher_driver_service):
        """Tests that a DriverError is raised for a non-existent driver."""
        with pytest.raises(DriverError, match="Pusher driver 'non_existent' not found."):
            pusher_driver_service._get_driver_by_type("non_existent")

    @pytest.mark.asyncio
    async def test_get_wizard_definition_by_type(self, pusher_driver_service, mock_pusher_class, mocker):
        """Tests that the service correctly proxies the call to get_wizard_steps."""
        mocker.patch.dict(pusher_driver_service._discovered_drivers, {"mock_pusher": mock_pusher_class})
        
        wizard = await pusher_driver_service.get_wizard_definition_by_type("mock_pusher")
        
        assert wizard == {"steps": [{"id": "s1"}]}
        mock_pusher_class.get_wizard_steps.assert_awaited_once()

    @pytest.mark.asyncio
    async def test_test_connection(self, pusher_driver_service, mock_pusher_class, mocker):
        """Tests that the service correctly proxies the call to test_connection."""
        mocker.patch.dict(pusher_driver_service._discovered_drivers, {"mock_pusher": mock_pusher_class})
        payload = {"endpoint": "http://test.com"}

        status, msg = await pusher_driver_service.test_connection("mock_pusher", **payload)

        assert status is True
        assert msg == "Connected"
        mock_pusher_class.test_connection.assert_awaited_once_with(**payload)

    @pytest.mark.asyncio
    async def test_check_privileges(self, pusher_driver_service, mock_pusher_class, mocker):
        """Tests that the service correctly proxies the call to check_privileges."""
        mocker.patch.dict(pusher_driver_service._discovered_drivers, {"mock_pusher": mock_pusher_class})
        payload = {"endpoint": "http://test.com", "credential": {"user": "u"}}

        status, msg = await pusher_driver_service.check_privileges("mock_pusher", **payload)

        assert status is True
        assert msg == "Privileges OK"
        mock_pusher_class.check_privileges.assert_awaited_once_with(**payload)

    @pytest.mark.asyncio
    async def test_get_needed_fields(self, pusher_driver_service, mock_pusher_class, mocker):
        """Tests that the service correctly proxies the call to get_needed_fields."""
        mocker.patch.dict(pusher_driver_service._discovered_drivers, {"mock_pusher": mock_pusher_class})
        payload = {"endpoint": "http://test.com"}

        fields = await pusher_driver_service.get_needed_fields("mock_pusher", **payload)

        assert fields == {"properties": {"field1": {"type": "string"}}}
        mock_pusher_class.get_needed_fields.assert_awaited_once_with(**payload)

    @pytest.mark.asyncio
    async def test_get_needed_fields_exception(self, pusher_driver_service, mock_pusher_class, mocker):
        """Tests that exceptions from the driver are caught and re-raised correctly."""
        mocker.patch.dict(pusher_driver_service._discovered_drivers, {"mock_pusher": mock_pusher_class})
        mock_pusher_class.get_needed_fields.side_effect = Exception("Test Exception")
        payload = {"endpoint": "http://test.com"}

        with pytest.raises(RuntimeError, match="Could not retrieve needed fields"):
            await pusher_driver_service.get_needed_fields("mock_pusher", **payload)


from .source import SourceConfigService
from .pusher import PusherConfigService
from .sync import SyncConfigService

__all__ = [
    "SourceConfigService", 
    "PusherConfigService", 
    "SyncConfigService"
]


from fastapi import APIRouter, Depends, Response
from typing import Dict, Any

from fustor_agent.services.instances.bus import EventBusService
from fustor_agent.services.instances.sync import SyncInstanceService
from fustor_agent.api.dependencies import get_app

router = APIRouter()

@router.get("/metrics", summary="Get Prometheus metrics", response_class=Response)
async def get_metrics(
    app = Depends(get_app)
):
    """
    Returns application metrics in Prometheus exposition format.
    """
    bus_service: EventBusService = app.bus_service
    sync_instance_service: SyncInstanceService = app.sync_instance_service

    metrics_output = []

    # Process Bus metrics
    for bus_id, bus_instance in bus_service.list_instances().items():
        bus_metrics = bus_instance.get_dto().statistics
        
        # fustor_agent_events_produced_total
        metrics_output.append(
            f'# HELP fustor_agent_events_produced_total Total number of events produced by source bus.\n'
            f'# TYPE fustor_agent_events_produced_total counter'
        )
        metrics_output.append(
            f'fustor_agent_events_produced_total{{source_id="{bus_instance.source_id}",bus_id="{bus_id}"}} {bus_metrics.get("events_produced_total", 0)}'
        )

        # fustor_agent_bus_queue_size
        metrics_output.append(
            f'# HELP fustor_agent_bus_queue_size Current number of events in the bus queue.\n'
            f'# TYPE fustor_agent_bus_queue_size gauge'
        )
        metrics_output.append(
            f'fustor_agent_bus_queue_size{{source_id="{bus_instance.source_id}",bus_id="{bus_id}"}} {bus_metrics.get("bus_queue_size", 0)}'
        )

    # Process Sync metrics
    for sync_id, sync_instance in sync_instance_service.list_instances().items():
        sync_metrics = sync_instance.get_dto().statistics

        # fustor_agent_events_pushed_total
        metrics_output.append(
            f'# HELP fustor_agent_events_pushed_total Total number of events successfully pushed to pusher.\n'
            f'# TYPE fustor_agent_events_pushed_total counter'
        )
        metrics_output.append(
            f'fustor_agent_events_pushed_total{{sync_id="{sync_id}",source_id="{sync_instance.config.source}",pusher_id="{sync_instance.config.pusher}"}} {sync_metrics.get("events_pushed_total", 0)}'
        )

        # fustor_agent_sync_push_latency_seconds
        # For a histogram, we'll just expose the sum and count for simplicity without a library
        # In a real scenario, you'd calculate buckets, but for a no-dependency approach, sum/count is a start.
        latencies = sync_metrics.get("sync_push_latency_seconds", [])
        if latencies:
            total_latency = sum(latencies)
            count_latency = len(latencies)
            metrics_output.append(
                f'# HELP fustor_agent_sync_push_latency_seconds Latency of pushing event batches to pushers.\n'
                f'# TYPE fustor_agent_sync_push_latency_seconds summary' # Using summary type for sum/count
            )
            metrics_output.append(
                f'fustor_agent_sync_push_latency_seconds_sum{{sync_id="{sync_id}",source_id="{sync_instance.config.source}",pusher_id="{sync_instance.config.pusher}"}} {total_latency}'
            )
            metrics_output.append(
                f'fustor_agent_sync_push_latency_seconds_count{{sync_id="{sync_id}",source_id="{sync_instance.config.source}",pusher_id="{sync_instance.config.pusher}"}} {count_latency}'
            )

    return Response(content="\n".join(metrics_output), media_type="text/plain; version=0.0.4")



import pytest
import os
import json
from unittest.mock import patch, mock_open

from fustor_agent.services import schema_cache

# Mock the SCHEMA_CACHE_DIR for testing purposes
@pytest.fixture(autouse=True)
def mock_schema_cache_dir():
    with patch('fustor_agent.services.schema_cache.SCHEMA_CACHE_DIR', '/tmp/test_schemas'):
        yield

@pytest.fixture
def sample_schema_data():
    return {
        "tables": {
            "users": {
                "columns": [
                    {"name": "id", "type": "int"},
                    {"name": "name", "type": "varchar"}
                ]
            }
        }
    }

class TestSchemaCache:
    @patch('os.makedirs')
    @patch('builtins.open', new_callable=mock_open)
    @patch('os.path.exists', side_effect=[False, True]) # First call for makedirs, second for file_path
    def test_save_source_schema_success(self, mock_exists, mock_file_open, mock_makedirs, sample_schema_data):
        source_id = "test_source"
        schema_cache.save_source_schema(source_id, sample_schema_data)
        
        mock_makedirs.assert_called_once_with('/tmp/test_schemas', exist_ok=True)
        mock_file_open.assert_called_once_with('/tmp/test_schemas/source_test_source.schema.json', 'w', encoding='utf-8')
        # Concatenate all calls to write and assert the final content
        written_content = "".join([call.args[0] for call in mock_file_open().write.call_args_list])
        assert written_content == json.dumps(sample_schema_data, indent=2)

    @patch('os.makedirs', side_effect=OSError("Permission denied"))
    @patch('os.path.exists', return_value=False)
    def test_save_source_schema_makedirs_failure(self, mock_exists, mock_makedirs, sample_schema_data):
        source_id = "test_source"
        with pytest.raises(OSError, match="Permission denied"):
            schema_cache.save_source_schema(source_id, sample_schema_data)

    @patch('os.makedirs')
    @patch('builtins.open', new_callable=mock_open)
    def test_save_source_schema_write_failure(self, mock_file_open, mock_makedirs, sample_schema_data):
        source_id = "test_source"
        mock_file_open.return_value.__enter__.return_value.write.side_effect = IOError("Disk full")
        with pytest.raises(IOError, match="Disk full"):
            schema_cache.save_source_schema(source_id, sample_schema_data)

    @patch('os.path.exists', return_value=True)
    @patch('builtins.open', new_callable=mock_open)
    def test_load_source_schema_success(self, mock_file_open, mock_exists, sample_schema_data):
        source_id = "test_source"
        mock_file_open.return_value.__enter__.return_value.read.return_value = json.dumps(sample_schema_data)
        
        loaded_schema = schema_cache.load_source_schema(source_id)
        assert loaded_schema == sample_schema_data
        mock_file_open.assert_called_once_with('/tmp/test_schemas/source_test_source.schema.json', 'r', encoding='utf-8')

    @patch('os.path.exists', return_value=False)
    def test_load_source_schema_file_not_found(self, mock_exists):
        source_id = "non_existent_source"
        loaded_schema = schema_cache.load_source_schema(source_id)
        assert loaded_schema is None

    @patch('os.path.exists', return_value=True)
    @patch('builtins.open', new_callable=mock_open)
    def test_load_source_schema_read_failure(self, mock_file_open, mock_exists):
        source_id = "test_source"
        mock_file_open.return_value.__enter__.return_value.read.side_effect = IOError("Read error")
        loaded_schema = schema_cache.load_source_schema(source_id)
        assert loaded_schema is None
    

    @patch('os.path.exists', return_value=True)
    @patch('builtins.open', new_callable=mock_open)
    def test_load_source_schema_json_decode_error(self, mock_file_open, mock_exists):
        source_id = "test_source"
        mock_file_open.return_value.__enter__.return_value.read.return_value = "invalid json"
        loaded_schema = schema_cache.load_source_schema(source_id)
        assert loaded_schema is None

    @patch('os.path.exists')
    def test_is_schema_valid_true(self, mock_exists):
        mock_exists.return_value = True
        source_id = "test_source"
        assert schema_cache.is_schema_valid(source_id) is True
        assert mock_exists.call_count == 2

    @patch('os.path.exists')
    def test_is_schema_valid_false_no_marker(self, mock_exists):
        mock_exists.side_effect = [True, False]
        source_id = "test_source"
        assert schema_cache.is_schema_valid(source_id) is False
        assert mock_exists.call_count == 2

    @patch('os.path.exists')
    def test_is_schema_valid_false_no_schema(self, mock_exists):
        mock_exists.return_value = False
        source_id = "non_existent_source"
        assert schema_cache.is_schema_valid(source_id) is False
        mock_exists.assert_called_once()



import asyncio
import logging
import queue
import threading
import time
from typing import Optional, Any, TYPE_CHECKING, Dict
from datetime import datetime, timezone

from fustor_core.models.config import SyncConfig, PusherConfig, SourceConfig
from fustor_core.models.states import SyncState, SyncInstanceDTO
from fustor_core.exceptions import DriverError
from fustor_agent.runtime.bus import EventBusFailedError

if TYPE_CHECKING:
    from fustor_agent.services.instances.bus import EventBusInstanceRuntime, EventBusService
    from fustor_agent.services.drivers.pusher_driver import PusherDriverService
    from fustor_agent.services.drivers.source_driver import SourceDriverService

logger = logging.getLogger("fustor_agent")

class SyncInstance:
    def __init__(
        self,
        id: str,
        agent_id: str,
        config: SyncConfig,
        source_config: SourceConfig,
        pusher_config: PusherConfig,
        bus_service: "EventBusService",
        pusher_driver_service: "PusherDriverService",
        source_driver_service: "SourceDriverService",
        pusher_schema: Dict[str, Any],
        initial_statistics: Optional[Dict[str, Any]] = None
    ):
        self.id = id
        self.task_id = f"{agent_id}:{id}"
        self.config = config
        self.source_config = source_config
        self.pusher_config = pusher_config
        self.bus_service = bus_service
        self.pusher_schema = pusher_schema
        self.session_id: Optional[str] = None
        
        source_driver_class = source_driver_service._get_driver_by_type(self.source_config.driver)
        self.source_driver_instance = source_driver_class(config.source, self.source_config)

        pusher_driver_class = pusher_driver_service._get_driver_by_type(self.pusher_config.driver)
        self.pusher_driver_instance = pusher_driver_class(config.pusher, self.pusher_config)

        self.bus: Optional["EventBusInstanceRuntime"] = None
        self.state: SyncState = SyncState.STOPPED
        self.info: str = "实例已创建, 等待启动。"
        self._main_task: Optional[asyncio.Task] = None
        self._statistics: Dict[str, Any] = initial_statistics if initial_statistics is not None else {"events_pushed": 0, "last_pushed_event_id": None}
        self.metrics: Dict[str, Any] = {"events_pushed_total": 0, "sync_push_latency_seconds": []}

        self._snapshot_task: Optional[asyncio.Task] = None
        
        # Heartbeat-related attributes
        self._heartbeat_task: Optional[asyncio.Task] = None
        self._stop_heartbeat_event = asyncio.Event()
        self._heartbeat_error_event = asyncio.Event()  # Event to signal heartbeat error
        self._last_active_time = datetime.now(timezone.utc)
        self.heartbeat_interval: int = 10  # Default heartbeat interval

        self._fast_mapper_fn = self._compile_mapper_function()

    def __str__(self):
        return f"Sync Instance {self.id}"
    
    def _set_state(self, new_state: SyncState, info: Optional[str] = None):
        # This method is for setting primary, exclusive states.
        # For adding/removing flags, directly manipulate self.state using bitwise operators.
        if self.state != new_state:
            logger.info(f"同步任务 '{self.id}' 状态变更: {self.state.name} -> {new_state.name}")
            self.state = new_state
        if info is not None:
            self.info = info

    async def start(self):
        if self.state != SyncState.STOPPED:
            logger.warning(f"任务 '{self.id}' 已在运行或处于非停止状态 ({self.state.name})，无法启动。")
            return

        if self._main_task and not self._main_task.done():
            self._main_task.cancel()
        
        self._main_task = asyncio.create_task(self._run_control_loop())
        
        # Start heartbeat task
        self._heartbeat_task = asyncio.create_task(self._run_heartbeat_loop())

    async def _run_heartbeat_loop(self):
        """后台心跳循环，定期更新锁状态"""
        while not self._stop_heartbeat_event.is_set():
            # Wait for session ID to be available before starting heartbeats
            while not self.session_id and not self._stop_heartbeat_event.is_set():
                await asyncio.sleep(0.5)  # Brief wait before checking again
            
            # If we're stopping, exit the loop
            if self._stop_heartbeat_event.is_set():
                break
                
            try:
                await self._send_heartbeat()
                
                await asyncio.wait_for(
                    self._stop_heartbeat_event.wait(), 
                    timeout=self.heartbeat_interval
                )
            except asyncio.TimeoutError:
                continue # Continue to next heartbeat
            except DriverError as e:
                logger.error(f"Heartbeat failed for sync '{self.id}' after multiple retries: {e}", exc_info=False)
                self._heartbeat_error_event.set()
                break # Exit loop on critical failure
            except Exception as e:
                logger.error(f"An unexpected error occurred in the heartbeat loop for sync '{self.id}': {e}", exc_info=True)
                self._heartbeat_error_event.set()
                break # Exit loop on critical failure

    async def _send_heartbeat(self):
        """发送心跳以维持会话状态"""
        result = await self.pusher_driver_instance.heartbeat(
            session_id=self.session_id
        )
        logger.debug(f"Heartbeat sent for sync '{self.id}', result: {result}")
        return result

    async def stop(self):
        if self.state == SyncState.STOPPED or SyncState.STOPPING in self.state:
            return
        
        # Only change state if we're not already in an error state
        if self.state != SyncState.ERROR:
            self._set_state(SyncState.STOPPING, "正在停止任务...")
        else:
            # If in error state, keep error state but perform cleanup
            logger.info(f"任务 '{self.id}' 处于错误状态，正在清理资源。")
        
        # 停止心跳任务
        if self._heartbeat_task:
            self._stop_heartbeat_event.set()
            self._heartbeat_task.cancel()
            try:
                await self._heartbeat_task
            except asyncio.CancelledError:
                pass
            except DriverError:
                # If heartbeat task ended with DriverError, it's already handled
                pass
        
        if self._main_task and not self._main_task.done():
            self._main_task.cancel()
        
        try:
            if self._main_task:
                await self._main_task
        except asyncio.CancelledError:
            pass

        # Close the driver's clients
        await self.pusher_driver_instance.close()
        
        # Also close the source driver to stop any background monitoring tasks
        if hasattr(self, 'source_driver_instance') and self.source_driver_instance:
            try:
                await self.source_driver_instance.close()
            except Exception as e:
                logger.warning(f"Error closing source driver instance: {e}")

        # Only set to STOPPED if not in error state
        if self.state != SyncState.ERROR:
            self._set_state(SyncState.STOPPED, "任务已停止。")
        else:
            logger.info(f"任务 '{self.id}' 保持错误状态，已停止相关任务。")

    async def _run_control_loop(self):
        try:
            # First, request a session from the Ingestor via the pusher driver
            session_data = await self.pusher_driver_instance.create_session(self.task_id)
            self.session_id = session_data.get("session_id")
            self.heartbeat_interval = session_data.get("suggested_heartbeat_interval_seconds", 10)
            
            logger.info(f"任务 '{self.id}' 正在启动，已从Ingestor获取会话 ID: {self.session_id}")
            
            self._set_state(SyncState.STARTING, "正在向接收端查询最新同步点位...")
            start_position = await self.pusher_driver_instance.get_latest_committed_index(session_id=self.session_id)

            # Set MESSAGE_SYNC state using bitwise OR
            self.state |= SyncState.MESSAGE_SYNC
            self.info = f"任务启动，进入消息同步阶段，起始点位: {start_position}"
            
            # Create tasks for monitoring heartbeat errors and message sync
            heartbeat_error_task = asyncio.create_task(self._monitor_heartbeat_errors())
            message_sync_task = asyncio.create_task(self._run_message_sync(start_position))
            
            # Wait for any of the tasks to complete
            done, pending = await asyncio.wait(
                [heartbeat_error_task, message_sync_task, self._heartbeat_task],
                return_when=asyncio.FIRST_COMPLETED
            )
            
            # Check if any completed task had an exception
            for task in done:
                if task.exception() is not None:
                    # If the heartbeat task had an exception, it means heartbeat failed
                    if task == self._heartbeat_task:
                        try:
                            await task  # This will raise the exception from the heartbeat task
                        except DriverError as e:
                            logger.error(f"Heartbeat task failed: {e}")
                            self._set_state(SyncState.ERROR, f"心跳失败: {e}")
                            # Set the heartbeat error event to notify other tasks
                            self._heartbeat_error_event.set()
                        except Exception as e:
                            logger.error(f"Heartbeat task failed with unexpected error: {e}")
                            self._set_state(SyncState.ERROR, f"心跳任务异常: {e}")
                            self._heartbeat_error_event.set()
                    else:
                        # Let the regular exception handling handle other exceptions
                        task.result()  # This will raise the exception if there was one

            # Cancel pending tasks
            for task in pending:
                task.cancel()
                try:
                    await task
                except asyncio.CancelledError:
                    pass

        except asyncio.CancelledError:
            logger.info(f"同步任务 '{self.id}' 被取消。")
        except DriverError as e:
            logger.error(f"同步任务 '{self.id}' 因驱动程序错误而失败: {e}")
            self._set_state(SyncState.ERROR, f"主控制循环崩溃: {e}")
        except EventBusFailedError as e:
            error_msg = f"数据总线错误: {e}"
            logger.error(f"同步任务 '{self.id}' 因底层事件总线错误而失败: {e}", exc_info=True)
            self._set_state(SyncState.ERROR, error_msg)
        except Exception as e:
            logger.error(f"同步任务 '{self.id}' 崩溃: {e}", exc_info=True)
            self._set_state(SyncState.ERROR, f"主控制循环崩溃: {e}")
        finally:
            # Cancel the heartbeat task when exiting the control loop
            if self._heartbeat_task and not self._heartbeat_task.done():
                self._heartbeat_task.cancel()
                try:
                    await self._heartbeat_task
                except asyncio.CancelledError:
                    pass
                except Exception:
                    # The heartbeat task might have already ended with an exception
                    pass

            if self._snapshot_task and not self._snapshot_task.done():
                self._snapshot_task.cancel()
                try:
                    await self._snapshot_task
                except asyncio.CancelledError:
                    pass
            # Ensure MESSAGE_SYNC is removed if it was set
            self.state &= ~SyncState.MESSAGE_SYNC
            logger.info(f"同步任务 '{self.id}' 结束。")

    async def _monitor_heartbeat_errors(self):
        """Monitor for heartbeat errors and set error state when they occur."""
        try:
            # Wait for the heartbeat error event to be set
            await self._heartbeat_error_event.wait()
            
            # When the event is set, it indicates a heartbeat failure that should stop the sync
            error_msg = f"心跳失败，同步任务停止。"
            self._set_state(SyncState.ERROR, error_msg)
            logger.error(f"Sync '{self.id}' stopped due to heartbeat failure: {error_msg}")
            
        except asyncio.CancelledError:
            # If the task is cancelled, just return
            pass

    async def _run_snapshot_sync(self):
        # Add SNAPSHOT_SYNC state using bitwise OR
        self.state |= SyncState.SNAPSHOT_SYNC
        self.info = "补充性质的快照同步任务开始运行。"
        snapshot_completed_successfully = False
        try:
            from fustor_agent.services.instances.bus import RequiredFieldsTracker
            
            event_queue = queue.Queue(maxsize=100)
            stop_event = threading.Event()

            def _threaded_snapshot_producer():
                try:
                    snapshot_field_tracker = RequiredFieldsTracker()
                    required_source_fields = {fm.source[0].split(':')[0] for fm in self.config.fields_mapping if fm.source}
                    snapshot_field_tracker.update_fields(required_source_fields)

                    iterator = self.source_driver_instance.get_snapshot_iterator(
                        batch_size=self.pusher_config.batch_size,
                        required_fields_tracker=snapshot_field_tracker
                    )
                    for event_batch in iterator:
                        if stop_event.is_set():
                            break
                        event_queue.put(event_batch)
                except Exception as e:
                    logger.error(f"Snapshot producer thread for '{self.id}' failed: {e}", exc_info=True)
                    event_queue.put(e)
                finally:
                    event_queue.put(None)

            producer_thread = threading.Thread(target=_threaded_snapshot_producer, daemon=True)
            producer_thread.start()

            while True:
                current_event = await asyncio.to_thread(event_queue.get)

                if current_event is None:
                    snapshot_completed_successfully = True # Set flag on natural exit
                    break
                
                if isinstance(current_event, Exception):
                    raise current_event

                if not current_event.rows:
                    continue

                final_rows_for_push = current_event.rows
                if self.config.fields_mapping:
                    mapped_rows = [self._process_field_mapping(row) for row in current_event.rows if self._process_field_mapping(row)]
                    if not mapped_rows:
                        continue
                    final_rows_for_push = mapped_rows
                
                current_event.rows = final_rows_for_push
                try:
                    push_start_time = time.monotonic()
                    await self.pusher_driver_instance.push(events=[current_event], session_id=self.session_id, source_type='snapshot')
                    push_duration = time.monotonic() - push_start_time
                    logger.info(f"Sync '{self.id}' snapshot push latency: {push_duration:.4f} seconds")
                except Exception as e:
                    # Check if this is the SessionObsoletedError specifically
                    if "Session is obsolete" in str(e):
                        logger.warning(f"Snapshot task for sync '{self.id}' is obsolete and was commanded to stop by the ingestor. Stopping gracefully.")
                        break
                    elif isinstance(e, DriverError) and ("419" in str(e) or "obsolete" in str(e).lower()):
                        logger.warning(f"Snapshot task for sync '{self.id}' is obsolete and was commanded to stop by the ingestor. Stopping gracefully.")
                        break
                    else:
                        raise
                self._statistics["events_pushed"] += len(current_event.rows)

            if snapshot_completed_successfully: # Only call if the loop finished naturally
                await self.pusher_driver_instance.push(events=[], task_id=self.id, session_id=self.session_id, is_snapshot_end=True, source_type='snapshot')
                logger.info(f"补充快照同步任务 '{self.id}' 完成。")

        except Exception as e:
            logger.error(f"补充快照同步任务 '{self.id}' 失败: {e}", exc_info=True)
        finally:
            stop_event.set()
            if producer_thread.is_alive():
                producer_thread.join(timeout=5.0)
            # Remove SNAPSHOT_SYNC state using bitwise AND with NOT
            self.state &= ~SyncState.SNAPSHOT_SYNC
            self.info = "快照同步任务已清理。"
            self._snapshot_task = None

    async def _run_message_sync(self, start_position: int):
        # Flag to track if we've sent the initial trigger event
        initial_event_sent = False
        
        try:
            self.bus, needed_position_lost = await self.bus_service.get_or_create_bus_for_subscriber(
                source_id=self.config.source,
                source_config=self.source_config,
                sync_id=self.id,
                required_position=start_position,
                fields_mapping=self.config.fields_mapping
            )
            if needed_position_lost:
                logger.warning(f"源 '{self.config.source}' 无法从请求的点位 {start_position} 开始，已从最新点位启动。")
                # 启动快照同步，因为位置丢失
                if not (self._snapshot_task and not self._snapshot_task.done()):
                    self._snapshot_task = asyncio.create_task(self._run_snapshot_sync())
                else:
                    logger.info(f"任务 '{self.id}' 已有一个快照正在运行，本次快照请求被忽略。")
        except DriverError as e:
            error_msg = f"无法启动事件总线或源驱动: {e}"
            self._set_state(SyncState.ERROR, error_msg)
            logger.error(error_msg, exc_info=True)
            return

        while SyncState.MESSAGE_SYNC in self.state:
            events_batch = await self.bus.internal_bus.get_events_for(
                self.id,
                self.pusher_config.batch_size,
                0.2
            )
            
            # If no events received and we haven't sent the initial trigger, create a fake event
            if not events_batch and not initial_event_sent:
                from fustor_event_model.models import UpdateEvent
                # Create a fake initial event to trigger the pusher and potentially start snapshot sync
                fake_event = UpdateEvent(
                    event_schema=self.config.source,  # Use source as event_schema
                    table="initial_trigger",    # Use a special table name for the trigger
                    rows=[],                    # Empty rows
                    index=start_position,       # Use the start position as the index
                    fields=[]                   # Empty fields for a fake event
                )
                events_batch = [fake_event]
                initial_event_sent = True
                logger.debug(f"Sent initial trigger event for task '{self.id}' to ensure pusher is called")
            
            if not events_batch: 
                continue
            
            events_to_push = []
            last_position_in_batch = 0
            for event in events_batch:
                if event.rows:
                    final_rows_for_push = event.rows
                    if self.config.fields_mapping:
                        mapped_rows = [self._process_field_mapping(row) for row in event.rows if self._process_field_mapping(row)]
                        if not mapped_rows:
                            last_position_in_batch = event.index
                            continue
                        final_rows_for_push = mapped_rows
                    
                    event.rows = final_rows_for_push
                events_to_push.append(event)
                
                last_position_in_batch = event.index

            if not events_to_push:
                await self.bus_service.commit_and_handle_split(
                    bus_id=self.bus.id, sync_id=self.id, num_events=len(events_batch),
                    last_consumed_position=last_position_in_batch, fields_mapping=self.config.fields_mapping
                )
                continue

            push_start_time = time.monotonic()
            response_dict = await self.pusher_driver_instance.push(
                events=events_to_push, 
                session_id=self.session_id,
                is_snapshot_end=False
            )
            push_duration = time.monotonic() - push_start_time
            logger.info(f"Sync '{self.id}' push latency: {push_duration:.4f} seconds")

            pushed_rows_count = sum(len(e.rows) for e in events_to_push)
            self._statistics["events_pushed"] += pushed_rows_count
            self._statistics["last_pushed_event_id"] = last_position_in_batch

            await self.bus_service.commit_and_handle_split(
                bus_id=self.bus.id, sync_id=self.id, num_events=len(events_batch),
                last_consumed_position=last_position_in_batch, fields_mapping=self.config.fields_mapping
            )

    def get_dto(self) -> SyncInstanceDTO:
        bus_info = self.bus.get_dto() if self.bus else None
        bus_id = self.bus.id if self.bus else None
        return SyncInstanceDTO(
            id=self.id,
            state=self.state,
            info=self.info,
            statistics=self._statistics,
            bus_info=bus_info,
            bus_id=bus_id
        )

    def _compile_mapper_function(self):
        if not self.config.fields_mapping:
            def passthrough_mapper(self, event_data, logger):
                return event_data
            import types
            logger.info(f"Sync '{self.id}' has no fields_mapping, using passthrough mapper.")
            return types.MethodType(passthrough_mapper, self)

        type_converter_map = {
            "string": "str",
            "integer": "int",
            "number": "float",
        }

        code_lines = [
            "def fast_mapper(self, event_data, logger):",
            "    processed_data = {}",
        ]

        endpoint_names = {m.to.split('.', 1)[0] for m in self.config.fields_mapping if '.' in m.to}
        for name in endpoint_names:
            code_lines.append(f"    processed_data['{name}'] = {{}}")

        for mapping in self.config.fields_mapping:
            if '.' not in mapping.to:
                continue

            endpoint_name, target_field_name = mapping.to.split('.', 1)
            
            target_field_schema = self.pusher_schema.get("properties", {}).get(f"{endpoint_name}.{target_field_name}")
            expected_type = target_field_schema.get("type") if target_field_schema else None

            if expected_type == "object":
                code_lines.append(f"    # Mapping for object: {mapping.to}")
                code_lines.append(f"    object_payload = {{}}")
                for source_str in mapping.source:
                    source_field_name = source_str.split(':')[0].split('.')[-1]
                    code_lines.append(f"    if '{source_field_name}' in event_data:")
                    code_lines.append(f"        object_payload['{source_field_name}'] = event_data.get('{source_field_name}')")
                code_lines.append(f"    processed_data['{endpoint_name}']['{target_field_name}'] = object_payload")
                continue

            if not mapping.source:
                continue
            
            source_field_name = mapping.source[0].split(':')[0].split('.')[-1]
            
            code_lines.append(f"    # Mapping for direct field: {mapping.to}")
            code_lines.append(f"    source_value = event_data.get('{source_field_name}')")
            code_lines.append(f"    if source_value is not None:")
            
            converter = type_converter_map.get(expected_type)
            if converter:
                code_lines.append(f"        try:")
                code_lines.append(f"            processed_data['{endpoint_name}']['{target_field_name}'] = {converter}(source_value)")
                code_lines.append(f"        except (ValueError, TypeError):")
                code_lines.append(f"            logger.warning(f'Failed to convert value {{source_value}} to type {expected_type} for field {mapping.to}')")
            else:
                code_lines.append(f"        processed_data['{endpoint_name}']['{target_field_name}'] = source_value")

        code_lines.append("    return processed_data")
        
        function_code = "\n".join(code_lines)
        
        local_namespace = {}
        try:
            exec(function_code, globals(), local_namespace)
            logger.info(f"Successfully compiled dynamic mapper function for sync '{self.id}'.")
            import types
            return types.MethodType(local_namespace['fast_mapper'], self)
        except Exception as e:
            logger.error(f"Failed to compile dynamic mapper function for sync '{self.id}': {e}", exc_info=True)
            return self._process_field_mapping_original

    def _process_field_mapping(self, event_data: Dict[str, Any]) -> Dict[str, Dict[str, Any]]:
        return self._fast_mapper_fn(event_data, logger)

    def _process_field_mapping_original(self, event_data: Dict[str, Any]) -> Dict[str, Dict[str, Any]]:
        logger.debug(f"[SyncID: {self.id}] --- Starting field mapping for event_data: {event_data}")
        processed_data = {}
        for mapping in self.config.fields_mapping:
            if '.' not in mapping.to: continue
            endpoint_name, target_field_name = mapping.to.split('.', 1)
            if endpoint_name not in processed_data: processed_data[endpoint_name] = {}
            target_field_schema = self.pusher_schema.get("properties", {}).get(f"{endpoint_name}.{target_field_name}")
            expected_type = target_field_schema.get("type") if target_field_schema else None
            if expected_type == "object":
                object_payload = {}
                for source_str in mapping.source:
                    parts = source_str.split(':')
                    if len(parts) != 2: continue
                    source_field_name_full = parts[0]
                    source_field_name = source_field_name_full.split('.')[-1]
                    source_value = event_data.get(source_field_name)
                    if source_value is not None:
                        object_key = source_field_name_full.split('.')[-1]
                        object_payload[object_key] = source_value
                processed_data[endpoint_name][target_field_name] = object_payload
            else:
                source_value = None
                for source_str in mapping.source:
                    source_field_name = source_str.split(':')[0].split('.')[-1]
                    if source_field_name in event_data:
                        source_value = event_data[source_field_name]
                        break
                if source_value is not None:
                    processed_data[endpoint_name][target_field_name] = source_value
        return processed_data

    async def remap_to_new_bus(self, new_bus: "EventBusInstanceRuntime", needed_position_lost: bool):
        self.bus = new_bus
        if needed_position_lost:
            logger.warning(f"Sync instance '{self.id}' was remapped to a new bus due to a split, and the new bus started from a later position. Starting snapshot sync...")
            # 启动快照同步，因为位置丢失
            if not (self._snapshot_task and not self._snapshot_task.done()):
                self._snapshot_task = asyncio.create_task(self._run_snapshot_sync())
            else:
                logger.info(f"任务 '{self.id}' 已有一个快照正在运行，本次快照请求被忽略。")
        logger.info(f"Sync instance '{self.id}' has been remapped to new bus '{new_bus.id}'.")



import pytest
from unittest.mock import MagicMock, AsyncMock, patch
from fustor_agent.services.configs.pusher import PusherConfigService
from fustor_core.models.config import AppConfig, PusherConfig, SyncConfig, PasswdCredential

@pytest.fixture
def mock_app_config():
    app_config = MagicMock(spec=AppConfig)
    app_config.get_pushers.return_value = {}
    app_config.get_syncs.return_value = {}
    return app_config

@pytest.fixture
def pusher_config_service(mock_app_config):
    service = PusherConfigService(mock_app_config)
    service.sync_instance_service = MagicMock() # Mock the injected dependency
    return service

@pytest.fixture
def sample_pusher_config():
    return PusherConfig(driver="http", endpoint="http://localhost", credential=PasswdCredential(user="u"), disabled=False)

@pytest.fixture
def sample_sync_config(sample_pusher_config):
    return SyncConfig(source="source1", pusher="pusher1", disabled=False)

class TestPusherConfigService:
    def test_set_dependencies(self, pusher_config_service):
        mock_sync_service = MagicMock()
        pusher_config_service.set_dependencies(mock_sync_service)
        assert pusher_config_service.sync_instance_service == mock_sync_service

    @pytest.mark.asyncio
    @patch('fustor_agent.services.configs.pusher.update_app_config_file')
    @patch('fustor_agent.services.common.config_lock')
    async def test_cleanup_obsolete_configs(self, mock_config_lock, mock_update_file, pusher_config_service, mock_app_config):
        # Setup mock app_config state
        initial_pushers = {
            "rec1": PusherConfig(driver="d", endpoint="e", credential=PasswdCredential(user="u"), disabled=True), # Obsolete
            "rec2": PusherConfig(driver="d", endpoint="e", credential=PasswdCredential(user="u"), disabled=False), # Not obsolete (enabled)
            "rec3": PusherConfig(driver="d", endpoint="e", credential=PasswdCredential(user="u"), disabled=True), # Obsolete
            "rec4": PusherConfig(driver="d", endpoint="e", credential=PasswdCredential(user="u"), disabled=False), # Not obsolete (enabled)
        }
        mock_app_config.get_pushers.return_value = initial_pushers
        mock_app_config.get_syncs.return_value = {
            "sync1": SyncConfig(source="s1", pusher="rec2", disabled=False),
            "sync2": SyncConfig(source="s1", pusher="rec4", disabled=False),
        }

        # Mock the async context manager for config_lock
        mock_config_lock.return_value.__aenter__ = AsyncMock(return_value=None)
        mock_config_lock.return_value.__aexit__ = AsyncMock(return_value=None)

        deleted_ids = await pusher_config_service.cleanup_obsolete_configs()

        assert sorted(deleted_ids) == sorted(["rec1", "rec3"])
        assert "rec1" not in mock_app_config.get_pushers.return_value
        assert "rec3" not in mock_app_config.get_pushers.return_value
        mock_update_file.assert_called_once()

    @pytest.mark.asyncio
    @patch('fustor_agent.services.configs.pusher.update_app_config_file')
    @patch('fustor_agent.services.common.config_lock')
    async def test_cleanup_obsolete_configs_no_obsolete(self, mock_config_lock, mock_update_file, pusher_config_service, mock_app_config):
        initial_pushers = {
            "rec1": PusherConfig(driver="d", endpoint="e", credential=PasswdCredential(user="u"), disabled=False),
            "rec2": PusherConfig(driver="d", endpoint="e", credential=PasswdCredential(user="u"), disabled=True), # Used by sync
        }
        mock_app_config.get_pushers.return_value = initial_pushers
        mock_app_config.get_syncs.return_value = {
            "sync1": SyncConfig(source="s1", pusher="rec2", disabled=False),
        }

        mock_config_lock.return_value.__aenter__ = AsyncMock(return_value=None)
        mock_config_lock.return_value.__aexit__ = AsyncMock(return_value=None)

        deleted_ids = await pusher_config_service.cleanup_obsolete_configs()

        assert deleted_ids == []
        mock_update_file.assert_not_called()



# src/fustor_agent/services/configs/pusher.py

import logging
from typing import Dict, Optional, List

from fustor_agent import update_app_config_file
from fustor_core.models.config import AppConfig, PusherConfig
from fustor_agent.services.instances.sync import SyncInstanceService
from fustor_agent.services.common import config_lock
from .base import BaseConfigService
from fustor_agent_sdk.interfaces import PusherConfigServiceInterface # Import the interface

logger = logging.getLogger("fustor_agent")

class PusherConfigService(BaseConfigService[PusherConfig], PusherConfigServiceInterface): # Inherit from the interface
    """
    Manages the lifecycle of PusherConfig objects.
    This service is responsible for CRUD operations on pusher configurations
    and inherits common logic from BaseConfigService.
    """
    def __init__(self, app_config: AppConfig):
        super().__init__(app_config, None, 'pusher')
        self.sync_instance_service: Optional[SyncInstanceService] = None

    def set_dependencies(self, sync_instance_service: SyncInstanceService):
        """Injects the SyncInstanceService for dependency management."""
        self.sync_instance_service = sync_instance_service

    # All common methods are inherited from BaseConfigService.

    async def cleanup_obsolete_configs(self) -> List[str]:
        """
        Finds and deletes all Pusher configurations that are both disabled and
        not used by any sync tasks.

        Returns:
            A list of the configuration IDs that were deleted.
        """
        all_sync_configs = self.app_config.get_syncs().values()
        in_use_pusher_ids = {sync.pusher for sync in all_sync_configs}

        all_pusher_configs = self.list_configs()
        obsolete_ids = [
            pusher_id for pusher_id, config in all_pusher_configs.items()
            if config.disabled and pusher_id not in in_use_pusher_ids
        ]

        if not obsolete_ids:
            logger.info("No obsolete pusher configurations to clean up.")
            return []

        logger.info(f"Found {len(obsolete_ids)} obsolete pusher configurations to clean up: {obsolete_ids}")

        deleted_ids = []
        async with config_lock:
            pusher_dict = self.app_config.get_pushers()
            for an_id in obsolete_ids:
                if an_id in pusher_dict:
                    pusher_dict.pop(an_id)
                    deleted_ids.append(an_id)
            
            if deleted_ids:
                update_app_config_file()
        
        logger.info(f"Successfully cleaned up {len(deleted_ids)} pusher configurations.")
        return deleted_ids


import logging
from fastapi import APIRouter, Depends, HTTPException, Body, status
from typing import List, Dict, Any, Optional
from pydantic import BaseModel

from fustor_agent.app import App
from fustor_agent.api.dependencies import get_app
from fustor_core.models.config import SourceConfig, PasswdCredential
from fustor_agent.services import schema_cache

from .schemas import ValidationResponse, CleanupResponse, ConfigCreateResponse

logger = logging.getLogger("fustor_agent")
router = APIRouter()


class AddSourcePayload(BaseModel):
    config: SourceConfig
    discovered_fields: Optional[Dict[str, Any]] = None


@router.post("/{id}", status_code=status.HTTP_201_CREATED, response_model=ConfigCreateResponse[SourceConfig], summary="Add a new Source configuration")
async def add_source_config(id: str, payload: AddSourcePayload, app: App = Depends(get_app)):
    """
    [REFACTORED] Adds a new Source configuration and caches its discovered fields if provided.
    This endpoint is now generic and relies on the client to provide the schema.
    """
    try:
        # Step 1: Add the main configuration.
        new_config = await app.source_config_service.add_config(id, payload.config)

        # Step 2: [UNIFIED LOGIC] If discovered fields were provided by the client,
        # cache them, regardless of the driver type.
        if payload.discovered_fields:
            schema_cache.save_source_schema(
                source_id=id,
                schema_data=payload.discovered_fields
            )
            # After saving, mark the schema as valid so the source can be enabled.
            schema_cache.validate_schema(id)

        return {"id": id, "config": new_config}
    except ValueError as e:
        raise HTTPException(status_code=status.HTTP_409_CONFLICT, detail=str(e))

@router.delete("/{id}", status_code=status.HTTP_204_NO_CONTENT, summary="Delete a Source configuration")
async def delete_source_config(id: str, app: App = Depends(get_app)):
    """Deletes a Source configuration.
    This will also stop any dependent sync tasks.""" 
    try:
        await app.source_config_service.delete_config(id)
    except ValueError as e:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=str(e))

@router.post("/_actions/cleanup", response_model=CleanupResponse, summary="Clean up obsolete Source configurations")
async def cleanup_obsolete_sources(app: App = Depends(get_app)):
    """Finds and deletes all disabled Source configurations that are not used by any sync task."""
    try:
        deleted_ids = await app.source_config_service.cleanup_obsolete_configs()
        count = len(deleted_ids) 
        return CleanupResponse(
            message=f"Cleanup successful. Deleted {count} obsolete source configurations.",
            deleted_count=count,
            deleted_ids=deleted_ids
        )
    except Exception as e:
        logger.error(f"Failed to cleanup obsolete source configs: {e}", exc_info=True)
        raise HTTPException(status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail="An unexpected error occurred during source cleanup.")

@router.post("/{id}/_actions/disable", status_code=status.HTTP_200_OK, summary="Disable a Source configuration")
async def disable_source_config(id: str, app: App = Depends(get_app)):
    try:
        await app.source_config_service.disable(id)
        return {"message": f"Source config '{id}' disabled successfully."}
    except ValueError as e:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=str(e)) 

@router.post("/{id}/_actions/enable", status_code=status.HTTP_200_OK, summary="Enable a Source configuration")
async def enable_source_config(id: str, app: App = Depends(get_app)):
    try:
        await app.source_config_service.enable(id)
        return {"message": f"Source config '{id}' enabled successfully."}
    except ValueError as e:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=str(e)) 

@router.post("/{id}/_actions/discover_and_cache_fields", summary="Discover and cache fields for a Source")
async def discover_and_cache_source_fields(id: str, admin_user: str = Body(..., embed=True), admin_password: str = Body(..., embed=True), app: App = Depends(get_app)):
    """
    Connects to a data source using provided admin credentials, discovers
    its available fields, and saves them to a local cache file.
    """ 
    try:
        source_config = app.source_config_service.get_config(id)
        if not source_config:
            raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=f"Source config '{id}' not found.")

        # Construct PasswdCredential from admin_user and admin_password
        admin_creds = PasswdCredential(user=admin_user, passwd=admin_password)

        await app.source_config_service.discover_and_cache_fields(
            id, admin_user, admin_password
        )
        # After successful discovery and caching, update the schema_cached status
        await app.source_config_service.update_schema_cached_status(id, True)
        return {"message": "Fields discovered and cached successfully."}
    except Exception as e:
        raise HTTPException(status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail=f"Failed to discover and cache fields for source '{id}': {e}") 

 

@router.get("/schema/{driver_type}", summary="Get the configuration schema for a specific Source driver")
async def get_source_driver_schema(driver_type: str, app: App = Depends(get_app)):
    try:
        schema = await app.source_driver_service.get_schema_by_type(driver_type)
        return schema
    except ValueError as e:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=str(e)) 
    except Exception as e:
        raise HTTPException(status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail=f"Failed to retrieve Source driver schema: {e}")


from fastapi import APIRouter, Depends
from fustor_agent.app import App
from .dependencies import get_app
from fustor_core.models.config import AppConfig

router = APIRouter()

@router.get("/", response_model=AppConfig, summary="Get the entire application configuration")
async def get_app_configuration(app: App = Depends(get_app)):
    return app._app_config


from typing import Dict, Optional, Any, List
from fustor_agent_sdk.interfaces import BaseInstanceServiceInterface # Import the interface

class BaseInstanceService(BaseInstanceServiceInterface): # Inherit from the interface
    """实例服务的抽象基类。"""
    def __init__(self):
        self.pool: Dict[str, Any] = {}

    def get_instance(self, id: str) -> Optional[Any]:
        """按ID查找一个运行时实例，找不到时返回None。"""
        return self.pool.get(id)

    def list_instances(self) -> List[Any]:
        """列出所有运行时实例。"""
        return list(self.pool.values())


import pytest
import asyncio
from fustor_agent.runtime.bus import MemoryEventBus, EventBusFailedError
from fustor_event_model.models import EventBase, InsertEvent
from fustor_core.models.config import FieldMapping

@pytest.mark.asyncio
class TestMemoryEventBus:
    async def test_put_and_get_events(self):
        bus = MemoryEventBus(bus_id="test_bus", capacity=10, start_position=0)
        event1 = InsertEvent(event_schema="s", table="t", rows=[{'id': 1}], fields=["id"], index=0)
        event2 = InsertEvent(event_schema="s", table="t", rows=[{'id': 2}], fields=["id"], index=1)

        await bus.subscribe("task1", 0, [])
        await bus.put(event1)
        await bus.put(event2)

        events = await bus.get_events_for("task1", 10, 0.1)
        assert len(events) == 2
        assert events[0].index == 0
        assert events[1].index == 1

    async def test_buffer_capacity(self):
        bus = MemoryEventBus(bus_id="test_bus", capacity=2, start_position=0)
        await bus.subscribe("task1", 0, [])
        await bus.put(InsertEvent(event_schema="s", table="t", rows=[{'id': 1}], fields=["id"], index=0))
        await bus.put(InsertEvent(event_schema="s", table="t", rows=[{'id': 2}], fields=["id"], index=1))

        put_task = asyncio.create_task(bus.put(InsertEvent(event_schema="s", table="t", rows=[{'id': 3}], fields=["id"], index=2)))
        
        await asyncio.sleep(0.01) # Give the put task time to block
        assert len(bus.buffer) == 2
        assert not put_task.done()

        await bus.commit("task1", 1, 0)
        await asyncio.sleep(0.01) # Give the put task time to unblock

        assert put_task.done()
        assert len(bus.buffer) == 2
        assert bus.buffer[0].index == 1
        assert bus.buffer[1].index == 2

    async def test_mark_as_failed(self):
        bus = MemoryEventBus(bus_id="test_bus", capacity=10, start_position=0)
        bus.mark_as_failed("Test failure")
        assert bus.failed is True
        with pytest.raises(EventBusFailedError):
            await bus.put(InsertEvent(event_schema="s", table="t", rows=[{'id': 1}], fields=["id"], index=0))

    async def test_subscribe_and_unsubscribe(self):
        bus = MemoryEventBus(bus_id="test_bus", capacity=10, start_position=0)
        assert bus.get_subscriber_count() == 0
        await bus.subscribe("task1", 0, [FieldMapping(to="a.b", source=["c.d.e:0"])])
        assert bus.get_subscriber_count() == 1
        assert "c.d.e" in bus.required_fields
        await bus.unsubscribe("task1")
        assert bus.get_subscriber_count() == 0
        assert not bus.required_fields

    async def test_subscribe_with_empty_fields_mapping_for_all_fields(self):
        """Tests that subscribing with an empty fields_mapping correctly sets required_fields to None (ALL_FIELDS)."""
        bus = MemoryEventBus(bus_id="test_bus_all_fields", capacity=10, start_position=0)
        
        # Subscribe with an empty fields_mapping
        await bus.subscribe("task_all", 0, [])
        
        # Assert that required_fields is None
        assert bus.required_fields is None
        
        # Test with another subscriber that has specific fields
        await bus.subscribe("task_specific", 0, [FieldMapping(to="a.b", source=["c.d.e:0"])])
        assert bus.required_fields is None # Still None because 'task_all' needs all

        await bus.unsubscribe("task_all")
        assert bus.required_fields == {"c.d.e"} # Now only specific fields are needed

        await bus.unsubscribe("task_specific")
        assert bus.required_fields == set() # No subscribers, so no fields needed

    async def test_trim_buffer(self):
        bus = MemoryEventBus(bus_id="test_bus", capacity=10, start_position=0)
        await bus.subscribe("task1", 0, [])
        await bus.subscribe("task2", 0, [])

        for i in range(5):
            await bus.put(InsertEvent(event_schema="s", table="t", rows=[{'id': i}], fields=["id"], index=i))
        
        assert len(bus.buffer) == 5
        await bus.commit("task1", 3, 2)
        assert len(bus.buffer) == 5 # No trim yet

        await bus.commit("task2", 2, 1)
        # low_watermark is now min(2, 1) = 1
        # Events with index <= 1 should be trimmed (0, 1)
        assert len(bus.buffer) == 3
        assert bus.buffer[0].index == 2

    async def test_check_for_split(self):
        bus = MemoryEventBus(bus_id="test_bus", capacity=10, start_position=0)
        await bus.subscribe("fast_consumer", 0, [])
        await bus.subscribe("slow_consumer", 0, [])

        for i in range(10):
            await bus.put(InsertEvent(event_schema="s", table="t", rows=[{'id': i}], fields=["id"], index=i))

        # Slow consumer stays at the start
        await bus.commit("slow_consumer", 1, 0)
        
        # Fast consumer moves to the end
        task_to_split = await bus.commit("fast_consumer", 8, 8)

        # 90% of capacity (9 events) are backlogged for the slow consumer
        assert task_to_split == "fast_consumer"

    async def test_can_subscribe(self):
        bus = MemoryEventBus(bus_id="test_bus", capacity=10, start_position=5)
        for i in range(5, 10):
            await bus.put(InsertEvent(event_schema="s", table="t", rows=[{'id': i}], fields=["id"], index=i))
        await bus.put(InsertEvent(event_schema="s", table="t", rows=[{'id': 5}], fields=["id"], index=5))

        assert bus.can_subscribe(4) is False
        assert bus.can_subscribe(5) is True
        assert bus.can_subscribe(6) is True
        assert bus.can_subscribe(7) is True

    async def test_update_subscriber_position(self):
        bus = MemoryEventBus(bus_id="test_bus", capacity=10, start_position=0)
        await bus.subscribe("task1", 0, [])
        await bus.update_subscriber_position("task1", 100)
        assert bus.subscribers["task1"]['last_consumed_index'] == 100



import pytest
from unittest.mock import MagicMock, AsyncMock, patch
from fustor_agent.services.configs.sync import SyncConfigService
from fustor_core.models.config import AppConfig, SyncConfig, PasswdCredential

@pytest.fixture
def mock_app_config():
    app_config = MagicMock(spec=AppConfig)
    app_config.get_syncs.return_value = {}
    return app_config

@pytest.fixture
def mock_source_config_service():
    return MagicMock()

@pytest.fixture
def mock_pusher_config_service():
    return MagicMock()

@pytest.fixture
def sync_config_service(mock_app_config, mock_source_config_service, mock_pusher_config_service):
    service = SyncConfigService(
        mock_app_config, 
        mock_source_config_service, 
        mock_pusher_config_service
    )
    service.sync_instance_service = MagicMock() # Mock the injected dependency
    return service

@pytest.fixture
def sample_sync_config():
    return SyncConfig(source="source1", pusher="pusher1", disabled=False)

class TestSyncConfigService:
    def test_set_dependencies(self, sync_config_service):
        mock_sync_service = MagicMock()
        sync_config_service.set_dependencies(mock_sync_service)
        assert sync_config_service.sync_instance_service == mock_sync_service

    # Inherits most of its functionality from BaseConfigService.
    # Additional tests can be added here if SyncConfigService introduces
    # unique logic beyond what BaseConfigService handles.



import pytest
from unittest.mock import MagicMock, AsyncMock, patch
from fustor_agent.services.configs.base import BaseConfigService
from fustor_core.models.config import AppConfig, SourceConfig, PusherConfig, SyncConfig, PasswdCredential
from fustor_core.exceptions import ConfigError, NotFoundError, ConflictError
from fustor_core.models.states import SyncState

# Define a simple mock config class for testing BaseConfigService
class MockConfig(SourceConfig):
    pass

@pytest.fixture
def mock_app_config():
    app_config = MagicMock(spec=AppConfig)
    app_config.get_sources.return_value = {}
    app_config.get_pushers.return_value = {}
    app_config.get_syncs.return_value = {}
    app_config.add_source = MagicMock()
    app_config.add_pusher = MagicMock()
    app_config.add_sync = MagicMock()
    app_config.delete_source = MagicMock()
    app_config.delete_pusher = MagicMock()
    app_config.delete_sync = MagicMock()
    return app_config

@pytest.fixture
def mock_sync_instance_service():
    service = MagicMock()
    service.mark_dependent_syncs_outdated = AsyncMock()
    service.stop_dependent_syncs = AsyncMock()
    service.stop_one = AsyncMock()
    service.get_instance = MagicMock()
    return service

@pytest.fixture
def base_config_service(mock_app_config, mock_sync_instance_service):
    return BaseConfigService(mock_app_config, mock_sync_instance_service, "source")

@pytest.fixture
def sample_source_config():
    return SourceConfig(driver="mysql", uri="mysql://host", credential=PasswdCredential(user="u"), disabled=False)

@pytest.fixture
def sample_pusher_config():
    return PusherConfig(driver="http", endpoint="http://localhost", credential=PasswdCredential(user="u"), disabled=False)

@pytest.fixture
def sample_sync_config():
    return SyncConfig(source="s1", pusher="r1", disabled=False)

class TestBaseConfigService:
    @pytest.mark.asyncio
    @patch('fustor_agent.services.configs.base.update_app_config_file')
    async def test_add_config(self, mock_update_file, base_config_service, mock_app_config, sample_source_config):
        mock_app_config.get_sources.return_value = {}
        mock_app_config.add_source.return_value = sample_source_config

        result = await base_config_service.add_config("test_source", sample_source_config)

        mock_app_config.add_source.assert_called_once_with("test_source", sample_source_config)
        mock_update_file.assert_called_once()
        assert result == sample_source_config

    @pytest.mark.asyncio
    @patch('fustor_agent.services.configs.base.update_app_config_file')
    async def test_update_config_enable_disable(self, mock_update_file, base_config_service, mock_app_config, mock_sync_instance_service, sample_source_config):
        # Setup initial state
        mock_app_config.get_sources.return_value = {"test_source": sample_source_config}

        # Test disabling
        updated_config = await base_config_service.update_config("test_source", {"disabled": True})
        assert updated_config.disabled is True
        mock_update_file.assert_called_once()
        mock_sync_instance_service.mark_dependent_syncs_outdated.assert_called_once_with(
            "source", "test_source", "Dependency Source 'test_source' configuration was disabled.", {"disabled": True}
        )
        mock_update_file.reset_mock()
        mock_sync_instance_service.mark_dependent_syncs_outdated.reset_mock()

        # Test enabling
        updated_config = await base_config_service.update_config("test_source", {"disabled": False})
        assert updated_config.disabled is False
        mock_update_file.assert_called_once()
        mock_sync_instance_service.mark_dependent_syncs_outdated.assert_called_once_with(
            "source", "test_source", "Dependency Source 'test_source' configuration was enabled.", {"disabled": False}
        )

    @pytest.mark.asyncio
    @patch('fustor_agent.services.configs.base.update_app_config_file')
    async def test_update_config_non_disabled_field(self, mock_update_file, base_config_service, mock_app_config, mock_sync_instance_service, sample_source_config):
        mock_app_config.get_sources.return_value = {"test_source": sample_source_config}

        updated_config = await base_config_service.update_config("test_source", {"max_retries": 20})
        assert updated_config.max_retries == 20
        mock_update_file.assert_called_once()
        mock_sync_instance_service.mark_dependent_syncs_outdated.assert_not_called()

    @pytest.mark.asyncio
    @patch('fustor_agent.services.configs.base.update_app_config_file')
    async def test_delete_config_source_pusher(self, mock_update_file, base_config_service, mock_app_config, mock_sync_instance_service, sample_source_config):
        mock_app_config.get_sources.return_value = {"test_source": sample_source_config}
        mock_app_config.delete_source.return_value = sample_source_config

        result = await base_config_service.delete_config("test_source")

        mock_app_config.delete_source.assert_called_once_with("test_source")
        mock_update_file.assert_called_once()
        assert result == sample_source_config

    @pytest.mark.asyncio
    @patch('fustor_agent.services.configs.base.update_app_config_file')
    async def test_delete_config_sync(self, mock_update_file, mock_app_config, mock_sync_instance_service, sample_sync_config):
        sync_service = BaseConfigService(mock_app_config, mock_sync_instance_service, "sync")
        mock_app_config.get_syncs.return_value = {"test_sync": sample_sync_config}
        mock_app_config.delete_sync.return_value = sample_sync_config

        result = await sync_service.delete_config("test_sync")

        mock_sync_instance_service.stop_one.assert_called_once_with("test_sync")
        mock_app_config.delete_sync.assert_called_once_with("test_sync")
        mock_update_file.assert_called_once()
        assert result == sample_sync_config

    @pytest.mark.asyncio
    async def test_disable_config(self, base_config_service, mock_app_config, sample_source_config):
        mock_app_config.get_sources.return_value = {"test_source": sample_source_config}
        with patch.object(base_config_service, 'update_config', new=AsyncMock()) as mock_update:
            await base_config_service.disable("test_source")
            mock_update.assert_called_once_with("test_source", {'disabled': True})

    @pytest.mark.asyncio
    async def test_enable_config(self, base_config_service, mock_app_config, sample_source_config):
        mock_app_config.get_sources.return_value = {"test_source": sample_source_config}
        with patch.object(base_config_service, 'update_config', new=AsyncMock()) as mock_update:
            await base_config_service.enable("test_source")
            mock_update.assert_called_once_with("test_source", {'disabled': False})

    @pytest.mark.asyncio
    async def test_delete_config_with_dependency(self, base_config_service, mock_app_config, sample_source_config, sample_sync_config):
        mock_app_config.get_sources.return_value = {"s1": sample_source_config}
        mock_app_config.get_syncs.return_value = {"sync1": sample_sync_config}

        with pytest.raises(ConflictError) as excinfo:
            await base_config_service.delete_config("s1")

        assert "used by the following sync tasks: sync1" in str(excinfo.value)

    @pytest.mark.asyncio
    async def test_update_config_sync_state_change(self, mock_app_config, mock_sync_instance_service, sample_sync_config):
        sync_service = BaseConfigService(mock_app_config, mock_sync_instance_service, "sync")
        mock_app_config.get_syncs.return_value = {"test_sync": sample_sync_config}

        mock_instance = MagicMock()
        # --- REFACTORED: Use a valid v2 state ---
        mock_instance.state = SyncState.MESSAGE_SYNC
        mock_instance._set_state = MagicMock()
        mock_sync_instance_service.get_instance.return_value = mock_instance

        await sync_service.update_config("test_sync", {"disabled": True})
        mock_instance._set_state.assert_called_once_with(SyncState.RUNNING_CONF_OUTDATE, "Dependency Sync 'test_sync' configuration was disabled.")

        mock_instance.state = SyncState.STOPPED
        mock_instance._set_state.reset_mock()
        await sync_service.update_config("test_sync", {"disabled": False})
        mock_instance._set_state.assert_not_called() # Should not set state if already stopped


import pytest
from fastapi import FastAPI
from fastapi.testclient import TestClient
from unittest.mock import MagicMock, AsyncMock, patch
from fustor_agent.app import App
from fustor_agent.api.routes import web_app
from fustor_core.models.config import AppConfig, PusherConfig, PasswdCredential, PusherConfigDict, SourceConfigDict, SyncConfigDict
from fustor_core.exceptions import ConfigError, NotFoundError

@pytest.fixture
def mock_app():
    app = MagicMock(spec=App)
    app._app_config = MagicMock(spec=AppConfig)
    app.pusher_config_service = MagicMock()
    app.pusher_config_service.add_config = AsyncMock()
    app.pusher_config_service.delete_config = AsyncMock()
    app.pusher_config_service.cleanup_obsolete_configs = AsyncMock(return_value=[])
    app.pusher_config_service.disable = AsyncMock()
    app.pusher_config_service.enable = AsyncMock()
    app.pusher_config_service.get_config = MagicMock()

    app.pusher_driver_service = MagicMock()
    app.pusher_driver_service.get_wizard_definition_by_type = AsyncMock()
    app.pusher_driver_service.test_connection = AsyncMock()
    app.pusher_driver_service.check_privileges = AsyncMock()
    app.pusher_driver_service.get_needed_fields = AsyncMock()

    return app

@pytest.fixture
def client(mock_app):
    from fustor_agent.api.dependencies import get_app
    from fustor_agent.app import App

    with patch.object(App, "startup", new=AsyncMock()):
        web_app.dependency_overrides[get_app] = lambda: mock_app
        with TestClient(web_app) as c:
            yield c
    web_app.dependency_overrides = {}

@pytest.fixture
def sample_pusher_config():
    return PusherConfig(driver="http", endpoint="http://localhost", credential=PasswdCredential(user="u"), disabled=False)

class TestPushersApi:
    def test_get_configs_includes_pushers(self, client, mock_app, sample_pusher_config):
        mock_app._app_config = AppConfig(
            sources=SourceConfigDict(root={}),
            pushers=PusherConfigDict(root={"test_pusher": sample_pusher_config}),
            syncs=SyncConfigDict(root={})
        )
        response = client.get("/api/configs/")
        assert response.status_code == 200
        data = response.json()
        assert isinstance(data, dict)
        assert "pushers" in data
        assert "test_pusher" in data["pushers"]

    @pytest.mark.asyncio
    async def test_add_pusher_config_success(self, client, mock_app, sample_pusher_config):
        mock_app.pusher_config_service.add_config.return_value = sample_pusher_config
        payload = {"config": sample_pusher_config.model_dump(mode='json')}
        response = client.post("/api/configs/pushers/new_pusher", json=payload)
        assert response.status_code == 201
        assert response.json() == {"id": "new_pusher", "config": sample_pusher_config.model_dump(mode='json')}
        mock_app.pusher_config_service.add_config.assert_called_once()

    @pytest.mark.asyncio
    async def test_add_pusher_config_conflict(self, client, mock_app, sample_pusher_config):
        mock_app.pusher_config_service.add_config.side_effect = ValueError("Pusher config already exists.")
        payload = {"config": sample_pusher_config.model_dump(mode='json')}
        response = client.post("/api/configs/pushers/existing_pusher", json=payload)
        assert response.status_code == 409
        assert "Pusher config already exists." in response.json()["detail"]

    @pytest.mark.asyncio
    async def test_delete_pusher_config_success(self, client, mock_app):
        response = client.delete("/api/configs/pushers/test_pusher")
        assert response.status_code == 204
        mock_app.pusher_config_service.delete_config.assert_called_once_with("test_pusher")

    @pytest.mark.asyncio
    async def test_delete_pusher_config_not_found(self, client, mock_app):
        mock_app.pusher_config_service.delete_config.side_effect = ValueError("Pusher config not found.")
        response = client.delete("/api/configs/pushers/non_existent")
        assert response.status_code == 404
        assert "Pusher config not found." in response.json()["detail"]

    @pytest.mark.asyncio
    async def test_cleanup_obsolete_pushers_success(self, client, mock_app):
        mock_app.pusher_config_service.cleanup_obsolete_configs.return_value = ["old_pusher1", "old_pusher2"]
        response = client.post("/api/configs/pushers/_actions/cleanup")
        assert response.status_code == 200
        assert response.json() == {
            "message": "Cleanup successful. Deleted 2 obsolete pusher configurations.",
            "deleted_count": 2,
            "deleted_ids": ["old_pusher1", "old_pusher2"]
        }
        mock_app.pusher_config_service.cleanup_obsolete_configs.assert_called_once()

    @pytest.mark.asyncio
    async def test_disable_pusher_config_success(self, client, mock_app):
        response = client.post("/api/configs/pushers/test_pusher/_actions/disable")
        assert response.status_code == 200
        assert response.json() == {"message": f"Pusher config 'test_pusher' disabled successfully."}
        mock_app.pusher_config_service.disable.assert_called_once_with("test_pusher")

    @pytest.mark.asyncio
    async def test_enable_pusher_config_success(self, client, mock_app):
        response = client.post("/api/configs/pushers/test_pusher/_actions/enable")
        assert response.status_code == 200
        assert response.json() == {"message": f"Pusher config 'test_pusher' enabled successfully."}
        mock_app.pusher_config_service.enable.assert_called_once_with("test_pusher")

    @pytest.mark.asyncio
    async def test_get_pusher_needed_fields(self, client, mock_app, sample_pusher_config):
        mock_app.pusher_config_service.get_config.return_value = sample_pusher_config
        mock_app.pusher_driver_service.get_needed_fields.return_value = {"field1": "type1"}
        # --- REFACTORED: Corrected the API path from '/configs' to '/drivers' ---
        response = client.get("/api/drivers/pushers/test_pusher/_actions/get_needed_fields")
        # --- END REFACTOR ---
        assert response.status_code == 200
        assert response.json() == {"field1": "type1"}
        mock_app.pusher_driver_service.get_needed_fields.assert_called_once_with("http", "http://localhost")

    @pytest.mark.asyncio
    async def test_get_pusher_driver_wizard(self, client, mock_app):
        mock_app.pusher_driver_service.get_wizard_definition_by_type.return_value = {"driver_param": "string"}
        response = client.get("/api/drivers/pushers/http/wizard")
        assert response.status_code == 200
        assert response.json() == {"driver_param": "string"}
        mock_app.pusher_driver_service.get_wizard_definition_by_type.assert_called_once_with("http")

    @pytest.mark.asyncio
    async def test_test_pusher_driver_connection(self, client, mock_app):
        mock_app.pusher_driver_service.test_connection.return_value = (True, "Connected")
        response = client.post("/api/drivers/pushers/http/_actions/test_connection", json={
            "endpoint": "http://test"
        })
        assert response.status_code == 200
        assert response.json() == {"success": True, "message": "Connected"}
        mock_app.pusher_driver_service.test_connection.assert_called_once()

    @pytest.mark.asyncio
    async def test_check_pusher_driver_privileges(self, client, mock_app):
        mock_app.pusher_driver_service.check_privileges.return_value = (True, "Privileges OK")
        response = client.post("/api/drivers/pushers/http/_actions/check_privileges", json={
            "endpoint": "http://test",
            "credential": {"user": "admin", "passwd": "pass"}
        })
        assert response.status_code == 200
        assert response.json() == {"success": True, "message": "Privileges OK"}
        mock_app.pusher_driver_service.check_privileges.assert_called_once()

    @pytest.mark.asyncio
    async def test_pusher_test_connection_saved_config(self, client, mock_app, sample_pusher_config):
        mock_app.pusher_config_service.get_config.return_value = sample_pusher_config
        mock_app.pusher_driver_service.test_connection.return_value = (True, "Connected")
        response = client.post("/api/configs/pushers/test_pusher/_actions/test_connection")
        assert response.status_code == 200
        assert response.json() == {"success": True, "message": "Connected"}
        mock_app.pusher_driver_service.test_connection.assert_called_once_with(driver_type='http', endpoint='http://localhost')

    @pytest.mark.asyncio
    async def test_pusher_check_privileges_saved_config(self, client, mock_app, sample_pusher_config):
        mock_app.pusher_config_service.get_config.return_value = sample_pusher_config
        mock_app.pusher_driver_service.check_privileges.return_value = (True, "Privileges OK")
        response = client.post("/api/configs/pushers/test_pusher/_actions/check_privileges")
        assert response.status_code == 200
        assert response.json() == {"success": True, "message": "Privileges OK"}
        mock_app.pusher_driver_service.check_privileges.assert_called_once_with(driver_type='http', endpoint='http://localhost', credential=sample_pusher_config.credential)


import logging
from typing import Optional, Dict, Any, List

from fustor_core.models.config import AppConfig, SyncConfig
from fustor_agent.services.instances.sync import SyncInstanceService
from .base import BaseConfigService
from .source import SourceConfigService
from .pusher import PusherConfigService
from fustor_agent_sdk.interfaces import SyncConfigServiceInterface # Import the interface

logger = logging.getLogger("fustor_agent")

class SyncConfigService(BaseConfigService[SyncConfig], SyncConfigServiceInterface): # Inherit from the interface
    """
    Manages SyncConfig objects.
    """
    def __init__(
        self,
        app_config: AppConfig,
        source_config_service: SourceConfigService,
        pusher_config_service: PusherConfigService
    ):
        super().__init__(app_config, None, 'sync')
        self.sync_instance_service: Optional[SyncInstanceService] = None
        self.source_config_service = source_config_service
        self.pusher_config_service = pusher_config_service

    def set_dependencies(self, sync_instance_service: SyncInstanceService):
        """Injects the SyncInstanceService for dependency management."""
        self.sync_instance_service = sync_instance_service

    async def enable(self, id: str):
        """Enables a Sync configuration, ensuring its source and pusher are also enabled."""
        # First, call the parent enable method to actually enable the sync config
        await super().enable(id)

        sync_config = self.get_config(id)
        if not sync_config:
            raise ValueError(f"Sync config '{id}' not found after enabling.") # Should not happen

        # Check if source is enabled
        source_config = self.source_config_service.get_config(sync_config.source)
        if not source_config:
            raise ValueError(f"Source '{sync_config.source}' for sync '{id}' not found.")
        if source_config.disabled:
            raise ValueError(f"Source '{sync_config.source}' for sync '{id}' is disabled. Please enable the source first.")

        # Check if pusher is enabled
        pusher_config = self.pusher_config_service.get_config(sync_config.pusher)
        if not pusher_config:
            raise ValueError(f"Pusher '{sync_config.pusher}' for sync '{id}' not found.")
        if pusher_config.disabled:
            raise ValueError(f"Pusher '{sync_config.pusher}' for sync '{id}' is disabled. Please enable the pusher first.")

        logger.info(f"Sync config '{id}' enabled successfully and its dependencies are active.")
        
    def get_wizard_definition(self) -> Dict[str, Any]:
        """
        Returns the step definitions for the Sync Task configuration wizard.
        This structure is fetched by the frontend to dynamically build the UI.
        """
        # Get lists of available (enabled) sources and pushers for dropdowns
        enabled_sources = [
            id for id, cfg in self.source_config_service.list_configs().items() if not cfg.disabled
        ]
        enabled_pushers = [
            id for id, cfg in self.pusher_config_service.list_configs().items() if not cfg.disabled
        ]

        return {
            "steps": [
                {
                    "step_id": "initial_selection",
                    "title": "选择源与目标",
                    "schema": {
                        "type": "object",
                        "properties": {
                            "id": {
                                "type": "string",
                                "title": "同步任务ID",
                                "description": "为新配置指定一个唯一的、易于识别的名称。"
                            },
                            "source": {
                                "type": "string",
                                "title": "选择 Source 配置",
                                "description": "选择一个已配置并启用的数据源。",
                                "enum": enabled_sources
                            },
                            "pusher": {
                                "type": "string",
                                "title": "选择 Pusher 配置",
                                "description": "选择一个已配置并启用的接收端。",
                                "enum": enabled_pushers
                            }
                        },
                        "required": ["id", "source", "pusher"]
                    },
                    "validations": [] # This step triggers a data load action, not a simple validation
                },
                {
                    "step_id": "field_mapping",
                    "title": "字段映射",
                    "schema": {
                        "type": "object",
                        "properties": {
                            "fields_mapping": {
                                "type": "array",
                                "title": "字段映射规则"
                            }
                        },
                        "description": "将Source提供的可用字段映射到Pusher需要的目标字段。名称相似的字段会被自动映射。"
                    },
                    "validations": [] # Validation is performed client-side (all required fields mapped)
                },
                {
                    "step_id": "advanced_settings",
                    "title": "高级参数",
                    "schema": {
                        "type": "object",
                        "properties": {
                        }
                    },
                    "validations": []
                }
            ]
        }


#!/usr/bin/env python
import click
import asyncio
import signal
import uvicorn
import os
import logging
import sys
import subprocess
import time

from fustor_common.logging_config import setup_logging
from fustor_common.paths import get_fustor_home_dir # NEW import
from . import CONFIG_DIR, CONFIG_FILE_NAME, ConfigurationError

# Define common logging path
HOME_FUSTOR_DIR = get_fustor_home_dir() # Use the common function
AGENT_LOG_FILE = os.path.join(HOME_FUSTOR_DIR, "agent.log") # Renamed from fustor_agent.log

# PID file location
PID_FILE = os.path.join(HOME_FUSTOR_DIR, "agent.pid") # Renamed from fustor_agent.pid


def _is_running():
    if not os.path.exists(PID_FILE):
        return False
    try:
        with open(PID_FILE, 'r') as f:
            pid = int(f.read().strip())
        os.kill(pid, 0)
    except (IOError, ValueError, OSError):
        try:
            os.remove(PID_FILE)
        except OSError:
            pass
        return False
    else:
        return pid

@click.group()
def cli():
    """FuAgent Command-Line Interface Tool"""
    pass

@cli.command()
@click.option("--reload", is_flag=True, help="Enable auto-reloading of the server on code changes (foreground only).")
@click.option("-p", "--port", default=8103, help="Port to run the server on.")
@click.option("-h", "--host", default="127.0.0.1", help="Host to bind the server to.")
@click.option("-D", "--daemon", is_flag=True, help="Run the service as a background daemon.")
@click.option("-V", "--verbose", is_flag=True, help="Enable verbose (DEBUG level) logging.")
@click.option("--no-console-log", is_flag=True, hidden=True, help="Internal: Disable console logging.")
def start(reload, port, host, daemon, verbose, no_console_log):
    """Starts the FuAgent monitoring service (in the foreground by default)."""
    log_level = "DEBUG" if verbose else "INFO"
    # Disable console logging if --no-console-log is passed (used by daemonized process)
    setup_logging(log_file_path=AGENT_LOG_FILE, base_logger_name="fustor_agent", level=log_level.upper(), console_output=(not no_console_log))
    logger = logging.getLogger("fustor_agent")

    if daemon:
        pid = _is_running()
        if pid:
            click.echo(f"FuAgent is already running with PID: {pid}")
            return
        
        click.echo("Starting FuAgent in the background...")
        # Use a common daemon launcher function to avoid module path issues
        import fustor_common.daemon as daemon_module
        daemon_module.start_daemon(
            service_module_path='fustor_agent.api.routes',
            app_var_name='web_app',
            pid_file_name='agent.pid',
            log_file_name='agent.log',
            display_name='FuAgent',
            port=port,
            host=host,  # Use the host parameter
            verbose=verbose,
            reload=reload  # Pass reload parameter
        )
        time.sleep(2) # Give the daemon time to start and write its PID file
        pid = _is_running()
        if pid:
            click.echo(f"FuAgent daemon started successfully with PID: {pid}")
        else:
            click.echo(click.style("Failed to start FuAgent daemon. Check logs for details.", fg="red"))
        return

    # --- Foreground Execution Logic ---
    if _is_running():
        click.echo("FuAgent is already running in the background. Stop it first.")
        return

    try:
        if not os.path.exists(CONFIG_DIR):
            os.makedirs(CONFIG_DIR)
        with open(PID_FILE, 'w') as f:
            f.write(str(os.getpid()))

        from .api.routes import web_app

        click.echo("\n" + "="*60)
        click.echo("FuAgent ")
        click.echo(f"Web : http://{host}:{port}/ui")
        click.echo("="*60 + "\n")
        
        app_to_run = web_app
        if reload:
            app_to_run = "fustor_agent.api.routes:web_app"

        # Configure uvicorn to use DEBUG level for access logs to reduce verbosity
        uvicorn_logger = logging.getLogger("uvicorn.access")
        uvicorn_logger.setLevel(logging.DEBUG)

        uvicorn.run(
            app_to_run,
            host=host,
            port=port,
            log_config=None,
            access_log=True,
            reload=reload,
        )

    except KeyboardInterrupt:
        click.echo("\nFuAgent is shutting down...")
    except ConfigurationError as e:
        click.echo("="*60)
        click.echo(click.style(f"FuAgent Configuration Error: {e}", fg="red"))
        click.echo(f"Please check your configuration file at: '{os.path.join(CONFIG_DIR, CONFIG_FILE_NAME)}'")
        click.echo("="*60)
    except Exception as e:
        logger.critical(f"An unexpected error occurred during startup: {e}", exc_info=True)
        click.echo(click.style(f"\nFATAL: An unexpected error occurred: {e}", fg="red"))
    finally:
        if os.path.exists(PID_FILE):
            os.remove(PID_FILE)

@cli.command()
def stop():
    """Stops the background FuAgent service."""
    pid = _is_running()
    if not pid:
        click.echo("FuAgent is not running.")
        return

    click.echo(f"Stopping FuAgent with PID: {pid}...")
    try:
        os.kill(pid, signal.SIGTERM)
        for _ in range(10):
            if not _is_running():
                break
            time.sleep(1)
        else:
            click.echo(click.style("FuAgent did not stop in time. Forcing shutdown.", fg="yellow"))
            os.kill(pid, signal.SIGKILL)

        click.echo("FuAgent stopped successfully.")
    except OSError as e:
        click.echo(click.style(f"Error stopping process: {e}", fg="red"))
    finally:
        if os.path.exists(PID_FILE):
            os.remove(PID_FILE)

@cli.command()
@click.option("--source-id", required=True, help="ID of the source configuration (e.g., 'mysql-prod').")
@click.option("--admin-user", required=False, default=None, help="Admin username for the source database (if required).")
@click.option("--admin-password", required=False, default=None, help="Admin password for the source database (if required).")
@click.pass_context
def discover_schema(ctx, source_id, admin_user, admin_password):
    """Discovers and caches the schema for a given source configuration."""
    # Setup with default INFO level for this one-off command
    setup_logging(log_file_path=AGENT_LOG_FILE, base_logger_name="fustor_agent", level="INFO")
    logger = logging.getLogger("fustor_agent")

    click.echo(f"Attempting to discover and cache schema for source: {source_id}...")
    try:
        from .app import App
        app_instance = App(config_dir=CONFIG_DIR)
        asyncio.run(app_instance.source_config_service.discover_and_cache_fields(
            source_id, admin_user, admin_password
        ))
        click.echo(f"Successfully discovered and cached schema for source '{source_id}'.")
    except ValueError as e:
        click.echo(f"Error: {e}", err=True)
        ctx.exit(1)
    except Exception as e:
        click.echo(f"An unexpected error occurred: {e}", err=True)
        logger.error(f"An unexpected error occurred while discovering schema for '{source_id}': {e}", exc_info=True)
        ctx.exit(1)


import asyncio
import logging
from collections import deque
from typing import Dict, List, Optional, Any, Set

from fustor_core.models.config import FieldMapping
from fustor_event_model.models import EventBase
from fustor_core.exceptions import TransientSourceBufferFullError

logger = logging.getLogger("fustor_agent")

# A type alias for clarity in subscriber state tracking
SubscriberState = Dict[str, Any]  # last_consumed_index: int, last_seen_position: int

class EventBusFailedError(Exception):
    """Custom exception raised when operating on a failed event bus."""
    pass

class MemoryEventBus:
    """
    An in-memory, multi-consumer event bus with automatic garbage collection
    and support for dynamic consumer splitting.
    """
    def __init__(self, bus_id: str, capacity: int, start_position: int):
        self.id = bus_id
        self.capacity = capacity
        self.buffer: deque[EventBase] = deque()
        self.subscribers: Dict[str, SubscriberState] = {}
        self.lock = asyncio.Lock()
        self._producer_can_put = asyncio.Event()
        self._producer_can_put.set()
        self._consumer_can_get = asyncio.Event()
        self.failed: bool = False
        self.error_message: Optional[str] = None
        self.buffer_start_position: int = start_position
        self.low_watermark: int = start_position
        self._next_event_index: int = start_position
        self.subscriber_field_map: Dict[str, Set[str]] = {}
        self.required_fields: Set[str] = set()

    def _recalculate_required_fields(self):
        if not self.subscriber_field_map: # If no subscribers, no fields are required
            self.required_fields = set()
            log_fields = set()
        else:
            all_fields: Optional[Set[str]] = None # Assume all fields needed if any subscriber needs all
            for fields_set in self.subscriber_field_map.values():
                if fields_set is None: # If any subscriber needs all fields, then the bus needs all fields
                    all_fields = None
                    break
                if all_fields is None: # Initialize if not already None
                    all_fields = set()
                all_fields.update(fields_set)
            self.required_fields = all_fields
            log_fields = "ALL_FIELDS" if all_fields is None else all_fields
        logger.debug(f"Bus '{self.id}': Recalculated required fields: {log_fields}")

    async def put(self, event: EventBase, is_transient: bool = False):
        if self.failed:
            raise EventBusFailedError(f"Bus '{self.id}' is in a failed state.")
        while True:
            async with self.lock:
                if len(self.buffer) < self.capacity:
                    break
                if is_transient:
                    raise TransientSourceBufferFullError(
                        f"Transient source's event buffer is filled up with no extra space left! (Capacity: {self.capacity})."
                        f"Consider doubling its size in the agent config, specifically at the max_queue_size parameter in the source config."
                    )
                self._producer_can_put.clear()
                logger.debug(f"Bus '{self.id}': Buffer full, producer is waiting.")
            await self._producer_can_put.wait()
            if self.failed:
                raise EventBusFailedError(f"Bus '{self.id}' failed while producer was waiting.")
        async with self.lock:
            last_index = self.buffer[-1].index if self.buffer else self.buffer_start_position - 1
            if event.index != -1:
                if event.index < last_index:
                    logger.warning(f"Bus '{self.id}': Discarding out-of-order event. Event index {event.index} < last index in buffer {last_index}.")
                    if len(self.buffer) < self.capacity:
                        self._producer_can_put.set()
                    return
                self._next_event_index = event.index + 1
            else:
                event.index = self._next_event_index
                self._next_event_index += 1
            self.buffer.append(event)
            logger.info(f"Bus '{self.id}': Put event with index {event.index}. Buffer size: {len(self.buffer)}/{self.capacity}")
            if len(self.buffer) < self.capacity:
                self._producer_can_put.set()
            self._consumer_can_get.set()

    async def get_events_for(self, sync_task_id: str, batch_size: int, timeout: float) -> List[EventBase]:
        if self.failed:
            raise EventBusFailedError(f"Bus '{self.id}' failed: {self.error_message}")
        async with self.lock:
            state = self.subscribers.get(sync_task_id)
            if not state: return []
            last_consumed_index = state.get('last_consumed_index', self.buffer_start_position - 1)
            events = []
            for event in self.buffer:
                if event.index > last_consumed_index:
                    events.append(event)
                    if len(events) >= batch_size:
                        break
            if events:
                return events
            self._consumer_can_get.clear()
        try:
            await asyncio.wait_for(self._consumer_can_get.wait(), timeout=timeout)
        except asyncio.TimeoutError:
            return []
        async with self.lock:
            state = self.subscribers.get(sync_task_id)
            if not state: return []
            last_consumed_index = state.get('last_consumed_index', self.buffer_start_position - 1)
            events = []
            for event in self.buffer:
                if event.index > last_consumed_index:
                    events.append(event)
                    if len(events) >= batch_size:
                        break
            return events

    async def commit(self, sync_task_id: str, num_events_consumed: int, last_consumed_position: int) -> Optional[str]:
        async with self.lock:
            state = self.subscribers.get(sync_task_id)
            if not state:
                logger.warning(f"Commit called for non-existent subscriber '{sync_task_id}' on bus '{self.id}'.")
                return None
            state['last_consumed_index'] = last_consumed_position
            state['last_seen_position'] = last_consumed_position
            self._update_low_watermark()
            task_to_split = self._check_for_split(committed_subscriber_id=sync_task_id)
            self._trim_buffer()
            return task_to_split

    def mark_as_failed(self, error: str):
        logger.error(f"Marking Bus '{self.id}' as FAILED. Reason: {error}")
        self.failed = True
        self.error_message = error
        self._producer_can_put.set()

    async def subscribe(self, sync_task_id: str, initial_position: int, fields_mapping: List[FieldMapping]):
        async with self.lock:
            if sync_task_id in self.subscribers:
                return
            self.subscribers[sync_task_id] = {
                'last_seen_position': initial_position - 1,
                'last_consumed_index': initial_position - 1
            }
            required_source_fields: Optional[Set[str]] = None # Use None to signify "all fields"
            if fields_mapping: # Only process fields_mapping if it's not empty
                required_source_fields = set()
                for fm in fields_mapping:
                    for source_str in fm.source:
                        field_name = source_str.split(':')[0]
                        required_source_fields.add(field_name)
            
            # Store the required fields. If None, it means all fields are required.
            self.subscriber_field_map[sync_task_id] = required_source_fields
            self._recalculate_required_fields()
            
            log_fields = "ALL_FIELDS" if required_source_fields is None else required_source_fields
            logger.info(f"Task '{sync_task_id}' subscribed to Bus '{self.id}' at event index {initial_position - 1}. Required fields: {log_fields}")
            self._update_low_watermark()

    async def unsubscribe(self, sync_task_id: str):
        async with self.lock:
            if sync_task_id in self.subscribers:
                del self.subscribers[sync_task_id]
                if sync_task_id in self.subscriber_field_map:
                    del self.subscriber_field_map[sync_task_id]
                self._recalculate_required_fields()
                logger.info(f"Task '{sync_task_id}' unsubscribed from Bus '{self.id}'.")
                self._update_low_watermark()
                
    def _update_low_watermark(self):
        if not self.subscribers:
            if self.buffer:
                self.low_watermark = self.buffer[-1].index
            return
        positions = [
            s.get('last_seen_position', self.buffer_start_position - 1)
            for s in self.subscribers.values()
        ]
        self.low_watermark = min(positions)

    def _trim_buffer(self):
        events_to_remove = 0
        for event in self.buffer:
            if event.index <= self.low_watermark:
                events_to_remove += 1
            else:
                break
        if events_to_remove > 0:
            logger.debug(f"Bus '{self.id}': trimming {events_to_remove} events older than or equal to watermark {self.low_watermark}.")
            for _ in range(events_to_remove):
                self.buffer.popleft()
            if self.buffer:
                self.buffer_start_position = self.buffer[0].index
        if len(self.buffer) < self.capacity:
            self._producer_can_put.set()
        
    def _check_for_split(self, committed_subscriber_id: str) -> Optional[str]:
        if len(self.subscribers) < 2:
            return None
        indices = {sid: s['last_consumed_index'] for sid, s in self.subscribers.items()}
        fastest_id = max(indices, key=lambda sid: indices[sid])
        if fastest_id != committed_subscriber_id:
            return None
        slowest_id = min(indices, key=lambda sid: indices[sid])
        slowest_consumer_index = indices[slowest_id]
        slowest_event_deque_position = -1
        for i, event in enumerate(self.buffer):
            if event.index == slowest_consumer_index:
                slowest_event_deque_position = i
                break
        if slowest_event_deque_position == -1:
            return None
        backlogged_events_count = (len(self.buffer) - 1) - slowest_event_deque_position
        if backlogged_events_count >= int(self.capacity * 0.95):
            logger.warning(
                f"Bus '{self.id}' split condition met! "
                f"Fastest task '{fastest_id}' is ahead of slowest task '{slowest_id}' "
                f"by approximately {backlogged_events_count} buffered events."
            )
            return fastest_id
        return None

    def can_subscribe(self, required_position: int) -> bool:
        if not self.buffer:
            return self.buffer_start_position == required_position
        return self.buffer_start_position <= required_position <= self.buffer[-1].index

    def get_subscriber_count(self) -> int:
        return len(self.subscribers)

    async def update_subscriber_position(self, sync_task_id: str, new_position: int):
        async with self.lock:
            state = self.subscribers.get(sync_task_id)
            if state:
                if new_position > state['last_consumed_index']:
                    state['last_consumed_index'] = new_position
                    state['last_seen_position'] = new_position
                    logger.info(f"Bus '{self.id}': Subscriber '{sync_task_id}' position updated to {new_position}.")
                    self._update_low_watermark()
            else:
                logger.warning(f"Bus '{self.id}': Attempted to update position for non-existent subscriber '{sync_task_id}'.")



import asyncio
from typing import TYPE_CHECKING, Dict, Any, Optional 
import logging 

from .base import BaseInstanceService
from fustor_core.models.states import SyncState
from fustor_agent.runtime.sync import SyncInstance
from fustor_core.exceptions import NotFoundError
from fustor_agent_sdk.interfaces import SyncInstanceServiceInterface # Import the interface

if TYPE_CHECKING:
    from fustor_agent.services.configs.sync import SyncConfigService
    from fustor_agent.services.configs.source import SourceConfigService
    from fustor_agent.services.configs.pusher import PusherConfigService
    from fustor_agent.services.instances.bus import EventBusService, EventBusInstanceRuntime
    from fustor_agent.services.drivers.pusher_driver import PusherDriverService
    from fustor_agent.services.drivers.source_driver import SourceDriverService

logger = logging.getLogger("fustor_agent")

class SyncInstanceService(BaseInstanceService, SyncInstanceServiceInterface): # Inherit from the interface
    def __init__(
        self, 
        sync_config_service: "SyncConfigService",
        source_config_service: "SourceConfigService",
        pusher_config_service: "PusherConfigService",
        bus_service: "EventBusService",
        pusher_driver_service: "PusherDriverService",
        source_driver_service: "SourceDriverService",
        agent_id: str  # Added agent_id
    ):
        super().__init__()
        self.sync_config_service = sync_config_service
        self.source_config_service = source_config_service
        self.pusher_config_service = pusher_config_service
        self.bus_service = bus_service
        self.pusher_driver_service = pusher_driver_service
        self.source_driver_service = source_driver_service
        self.agent_id = agent_id  # Store agent_id
        self.logger = logging.getLogger("fustor_agent") 

    async def start_one(self, id: str):
        self.logger.debug(f"Enter start_one for sync_id: {id}")

        if self.get_instance(id):
            self.logger.warning(f"Sync instance '{id}' is already running or being managed.")
            return

        sync_config = self.sync_config_service.get_config(id)
        if not sync_config:
            raise NotFoundError(f"Sync config '{id}' not found.")

        if sync_config.disabled:
            self.logger.info(f"Sync instance '{id}' will not be started because its configuration is disabled.")
            return

        source_config = self.source_config_service.get_config(sync_config.source)
        if not source_config:
            raise NotFoundError(f"Source config '{sync_config.source}' not found for sync '{id}'.")
        
        pusher_config = self.pusher_config_service.get_config(sync_config.pusher)
        if not pusher_config:
            raise NotFoundError(f"Required Pusher config '{sync_config.pusher}' not found.")
        
        self.logger.info(f"Attempting to start sync instance '{id}'...")
        try:
            pusher_schema = await self.pusher_driver_service.get_needed_fields(
                driver_type=pusher_config.driver, endpoint=pusher_config.endpoint
            )
            # Instantiate the new, more complex SyncInstance, passing all dependencies.
            sync_instance = SyncInstance(
                id=id,
                agent_id=self.agent_id, # Pass agent_id
                config=sync_config,
                source_config=source_config,
                pusher_config=pusher_config,
                bus_service=self.bus_service,
                pusher_driver_service=self.pusher_driver_service,
                source_driver_service=self.source_driver_service,
                pusher_schema=pusher_schema
            )
            self.pool[id] = sync_instance
            await sync_instance.start()
            
            self.logger.info(f"Sync instance '{id}' start initiated successfully.")

        except Exception as e:
            self.logger.error(f"Failed to start sync instance '{id}': {e}", exc_info=True)
            if self.get_instance(id):
                self.pool.pop(id)
            # Re-raise to be caught by the API layer
            raise

    async def stop_one(self, id: str, should_release_bus: bool = True):
        instance = self.get_instance(id)
        if not instance:
            return

        self.logger.info(f"Attempting to stop {instance}...")
        try:
            # The instance's bus might not exist if stopped during snapshot phase
            bus_id = instance.bus.id if instance.bus else None
            
            await instance.stop()
            self.pool.pop(id, None)
            self.logger.info(f"{instance} stopped and removed from pool.")
            
            if should_release_bus and bus_id:
                await self.bus_service.release_subscriber(bus_id, id)
        except Exception as e:
            self.logger.error(f"Failed to cleanly stop {instance}: {e}", exc_info=True)

    async def remap_sync_to_new_bus(self, sync_id: str, new_bus: "EventBusInstanceRuntime", needed_position_lost: bool):
        sync_instance = self.get_instance(sync_id)
        if not sync_instance:
            self.logger.warning(f"Sync task '{sync_id}' not found during bus remapping.")
            return

        old_bus_id = sync_instance.bus.id if sync_instance.bus else None
        self.logger.info(f"Remapping sync task '{sync_id}' to new bus '{new_bus.id}'...")
        
        # Call the instance's remap method, which also handles the signal
        await sync_instance.remap_to_new_bus(new_bus, needed_position_lost)
        
        if old_bus_id:
            await self.bus_service.release_subscriber(old_bus_id, sync_id)
        
        self.logger.info(f"Sync task '{sync_id}' remapped to bus '{new_bus.id}' successfully.")

    async def mark_dependent_syncs_outdated(self, dependency_type: str, dependency_id: str, reason_info: str, updates: Optional[Dict[str, Any]] = None):
        affected_syncs = [
            inst for inst in self.list_instances() 
            if (dependency_type == "source" and inst.config.source == dependency_id) or
               (dependency_type == "pusher" and inst.config.pusher == dependency_id)
        ]

        logger.info(f"Marking syncs dependent on {dependency_type} '{dependency_id}' as outdated.")
        for sync_instance in affected_syncs:
            # Add RUNNING_CONF_OUTDATE flag using bitwise OR
            # This allows it to stack with existing states like SNAPSHOT_SYNC or MESSAGE_SYNC
            sync_instance.state |= SyncState.RUNNING_CONF_OUTDATE
            sync_instance.info = reason_info # Update info directly
        self.logger.info(f"Marked {len(affected_syncs)} syncs as outdated.")

    async def start_all_enabled(self):
        all_sync_configs = self.sync_config_service.list_configs()
        if not all_sync_configs:
            return

        self.logger.info(f"Attempting to auto-start all enabled sync tasks...")
        start_tasks = [
            self.start_one(id)
            for id, cfg in all_sync_configs.items() if not cfg.disabled
        ]
        if start_tasks:
            # We must run tasks sequentially to avoid race conditions if they share resources,
            # but for now, gather is acceptable for starting independent tasks.
            await asyncio.gather(*start_tasks, return_exceptions=True)
            
    async def restart_outdated_syncs(self) -> int:
        outdated_syncs = [
            inst for inst in self.list_instances() 
            if inst.state == SyncState.RUNNING_CONF_OUTDATE
        ]
        
        if not outdated_syncs:
            return 0
            
        self.logger.info(f"Found {len(outdated_syncs)} outdated sync tasks to restart.")
        
        for sync in outdated_syncs:
            await self.stop_one(sync.id) 
            await asyncio.sleep(1) # Give time for graceful shutdown
            await self.start_one(sync.id)
        
        return len(outdated_syncs)

    async def stop_all(self):
        self.logger.info("Stopping all sync instances...")
        keys_to_stop = list(self.pool.keys())
        stop_tasks = [self.stop_one(key, should_release_bus=False) for key in keys_to_stop]
        
        if stop_tasks:
            await asyncio.gather(*stop_tasks)

        await self.bus_service.release_all_unused_buses()


import logging
from fastapi import APIRouter, Depends, Body, HTTPException, status
from typing import Dict, List, Any, Optional, Union
from pydantic import BaseModel

from fustor_agent.app import App
from .dependencies import get_app
from .schemas import MessageResponse, ValidationResponse, TestSourceConnectionRequest, TestPusherConnectionRequest, AdminCredentials
from fustor_core.models.config import Credential, PasswdCredential, SourceConfig
from fustor_agent.services import schema_cache
from fustor_core.exceptions import DriverError, ConfigError

logger = logging.getLogger("fustor_agent")
router = APIRouter()

class WizardUserActionPayload(BaseModel):
    uri: str
    admin_creds: Credential
    agent_creds: Credential

@router.get("/", summary="List all available drivers")
async def list_available_drivers(app: App = Depends(get_app)) -> Dict[str, List[str]]:
    source_drivers = app.source_driver_service.list_available_drivers()
    pusher_drivers = app.pusher_driver_service.list_available_drivers()
    return {"sources": source_drivers, "pushers": pusher_drivers}

@router.get("/sources/{driver_type}/wizard", summary="Get the wizard definition for a specific Source driver")
async def get_source_driver_wizard(driver_type: str, app: App = Depends(get_app)) -> Dict[str, Any]:
    try:
        wizard_definition = await app.source_driver_service.get_wizard_definition_by_type(driver_type)
        return wizard_definition
    except (DriverError, ConfigError) as e:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=str(e))
    except Exception as e:
        logger.error(f"Failed to retrieve wizard definition for driver '{driver_type}': {e}", exc_info=True)
        raise HTTPException(status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail="An unexpected error occurred while fetching the wizard definition.")

@router.get("/pushers/{driver_type}/wizard", summary="Get the wizard definition for a specific Pusher driver")
async def get_pusher_driver_wizard(driver_type: str, app: App = Depends(get_app)) -> Dict[str, Any]:
    try:
        wizard_definition = await app.pusher_driver_service.get_wizard_definition_by_type(driver_type)
        return wizard_definition
    except (DriverError, ConfigError) as e:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=str(e))
    except Exception as e:
        logger.error(f"Failed to retrieve wizard definition for pusher driver '{driver_type}': {e}", exc_info=True)
        raise HTTPException(status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail="An unexpected error occurred while fetching the wizard definition.")

@router.get("/sources/{id}/_actions/get_available_fields", summary="Get cached available fields for a Source")
async def get_source_available_fields(id: str, app: App = Depends(get_app)):
    fields = schema_cache.load_source_schema(id)
    if fields is None:
        return {"properties": {}}
    return fields

@router.get("/pushers/{id}/_actions/get_needed_fields", summary="Get all needed fields from a configured Pusher")
async def get_pusher_needed_fields(id: str, app: App = Depends(get_app)):
    try:
        pusher_config = app.pusher_config_service.get_config(id)
        if not pusher_config:
            raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=f"Pusher config '{id}' not found.")
        fields = await app.pusher_driver_service.get_needed_fields(
            pusher_config.driver, pusher_config.endpoint
        )
        return fields
    except Exception as e:
        raise HTTPException(status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail=f"Failed to retrieve needed fields from pusher '{id}': {e}")

@router.post("/sources/{driver_type}/_actions/test_connection", response_model=ValidationResponse, summary="Test connection for a source driver")
async def test_source_connection(driver_type: str, payload: Dict[str, Any] = Body(...), app: App = Depends(get_app)):
    try:
        ok, msg = await app.source_driver_service.test_connection(driver_type=driver_type, **payload)
        if ok:
            return ValidationResponse(success=True, message=msg)
        else:
            raise ValueError(msg)
    except Exception as e:
        logger.error(f"Failed to test source connection for {driver_type}: {e}")
        raise HTTPException(status_code=400, detail=str(e))

@router.post("/sources/{driver_type}/_actions/check_params", response_model=ValidationResponse, summary="Check runtime parameters for a given Source driver")
async def check_source_params(driver_type: str, payload: Dict[str, Any] = Body(...), app: App = Depends(get_app)):
    try:
        ok, msg = await app.source_driver_service.check_params(driver_type, **payload)
        return ValidationResponse(success=ok, message=msg)
    except Exception as e:
        logger.error(f"Error checking parameters for driver '{driver_type}': {e}", exc_info=True)
        raise HTTPException(status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail=f"An unexpected error occurred: {e}")

@router.post("/sources/{driver_type}/_actions/discover_fields_no_cache", summary="Discover fields for a driver without caching")
async def discover_source_fields_no_cache(driver_type: str, payload: Dict[str, Any] = Body(...), app: App = Depends(get_app)):
    try:
        fields = await app.source_driver_service.get_available_fields(driver_type, **payload)
        field_count = len(fields.get("properties", {}))
        return {
            "success": True,
            "message": f"成功发现 {field_count} 个可用字段。",
            "fields": fields
        }
    except Exception as e:
        logger.error(f"Error during field discovery for driver '{driver_type}': {e}", exc_info=True)
        if isinstance(e, DriverError):
            raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail=str(e))
        raise HTTPException(status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail=f"An unexpected error occurred: {e}")

@router.post("/sources/{driver_type}/_actions/create_agent_user", response_model=ValidationResponse, summary="Create an agent user without a saved config")
async def create_source_user(driver_type: str, payload: Dict[str, Any] = Body(...), app: App = Depends(get_app)):
    try:
        ok, msg = await app.source_driver_service.create_agent_user(driver_type, **payload)
        return ValidationResponse(success=ok, message=msg)
    except Exception as e:
        logger.error(f"Error during create_agent_user for driver '{driver_type}': {e}", exc_info=True)
        raise HTTPException(status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail=f"An unexpected error occurred: {e}")

@router.post("/sources/{driver_type}/_actions/check_privileges", response_model=ValidationResponse, summary="Check agent user privileges without a saved config")
async def check_source_privileges(driver_type: str, payload: Dict[str, Any], app: App = Depends(get_app)):
    ok, msg = await app.source_driver_service.check_privileges(driver_type, **payload)
    return ValidationResponse(success=ok, message=msg)

@router.post("/pushers/{driver_type}/_actions/test_connection", response_model=ValidationResponse, summary="Test connection for a pusher driver")
async def test_pusher_connection(driver_type: str, payload: Dict[str, Any] = Body(...), app: App = Depends(get_app)):
    try:
        endpoint = payload.get("endpoint")
        if not endpoint:
            raise HTTPException(status_code=400, detail="Missing 'endpoint' in request body.")
        
        # Pass the whole payload to the service layer
        ok, msg = await app.pusher_driver_service.test_connection(
            driver_type=driver_type,
            **payload
        )
        if ok:
            return ValidationResponse(success=True, message=msg or "Pusher connection successful.")
        else:
            # Create a ValidationResponse for failure case for consistency
            return ValidationResponse(success=False, message=msg or "Pusher connection failed.")
    except Exception as e:
        logger.error(f"Failed to test pusher connection for {driver_type}: {e}", exc_info=True)
        # Return a standard validation response on exception
        return ValidationResponse(success=False, message=f"An exception occurred: {e}")

@router.post("/pushers/{driver_type}/_actions/check_privileges", response_model=ValidationResponse, summary="Check privileges for a given Pusher driver")
async def check_pusher_privileges(driver_type: str, payload: Dict[str, Any] = Body(...), app: App = Depends(get_app)):
    try:
        ok, msg = await app.pusher_driver_service.check_privileges(
            driver_type,
            **payload
        )
        return ValidationResponse(success=ok, message=msg)
    except Exception as e:
        logger.error(f"Error checking privileges for driver '{driver_type}': {e}", exc_info=True)
        return ValidationResponse(success=False, message=f"An exception occurred: {e}")

# --- START: NEW ENDPOINT for Pusher field discovery in wizard ---
@router.post("/pushers/{driver_type}/_actions/discover_fields", summary="Discover needed fields for a Pusher driver configuration")
async def discover_pusher_fields(driver_type: str, payload: Dict[str, Any] = Body(...), app: App = Depends(get_app)):
    """
    Connects to the pusher's endpoint (e.g., openapi.json) and discovers the
    fields it requires. This is used as a validation step in the UI wizard.
    """
    try:
        endpoint = payload.get("endpoint")
        if not endpoint:
            raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail="Payload must include 'endpoint'.")

        fields = await app.pusher_driver_service.get_needed_fields(driver_type, endpoint)
        field_count = len(fields.get("properties", {}))
        return {
            "success": True,
            "message": f"成功发现 {field_count} 个目标字段。",
            "fields": fields
        }
    except Exception as e:
        logger.error(f"Error discovering pusher fields for driver '{driver_type}': {e}", exc_info=True)
        # Return a standard validation response on error
        return ValidationResponse(success=False, message=f"发现字段时出错: {e}")
# --- END: NEW ENDPOINT ---